ALERT_DIALOG_DESCRIPTION = '//android.widget.TextView[@resource-id="com.snapchat.android:id/alert_dialog_description"]'
CAMERA_CAPTURE = '//android.widget.FrameLayout[@content-desc="Camera Capture"]'
CAMERA_PAGE = '//android.widget.FrameLayout[@resource-id="com.snapchat.android:id/camera_page"]'
CAPTION_EDIT = '//android.widget.EditText[@resource-id="com.snapchat.android:id/caption_edit_text_view"]'
CHAT_INPUT = '//android.widget.EditText[@resource-id="com.snapchat.android:id/chat_input_text_field"]'
CHAT_CONVERSATION = ('//android.view.View['
                     '@resource-id="com.snapchat.android:id/ff_item"'
                     ' and .//javaClass[@text="{conversation}"]]')
CHAT_TAB = '//android.view.ViewGroup[@content-desc="Chat"]'
CONVERSATION_TITLE = '//android.widget.TextView[@resource-id="com.snapchat.android:id/conversation_title_text_view"]'
CONVERSATION_TITLE_TEXT = '//android.widget.TextView[@resource-id="com.snapchat.android:id/conversation_title_text_view" and contains(lower-case(@text), "{conversation}")]'
DISCARD_BUTTON = '//android.widget.LinearLayout[@resource-id="com.snapchat.android:id/dialog_content"]/android.view.View'
FEED_NEW_CHAT = '//android.widget.FrameLayout[@resource-id="com.snapchat.android:id/feed_new_chat"]'
FULL_SCREEN_SURFACE_VIEW = '//android.view.SurfaceView[@resource-id="com.snapchat.android:id/full_screen_surface_view"]'
RECIPIENTS_TO_ADD = ('//androidx.recyclerview.widget.RecyclerView[@resource-id="com.snapchat.android:id/send_to_recycler_view"]'
                  '//android.view.View[count(.//javaClass)=1]//javaClass[@text="{recipient}"]')
MY_STORY = '//javaClass[@text="My Story · Friends Only"]/..'
NEW_STORY = '//android.view.View[@content-desc="New Story Button"]'
SEND_BUTTON = '//android.view.View[@content-desc="Send"]'
SENT_TO = '//android.view.ViewGroup[@resource-id="com.snapchat.android:id/sent_to_button_label_mode_view"]'
TOGGLE_CAMERA = '(//android.widget.ImageView[@resource-id="com.snapchat.android:id/camera_mode_icon_image_view"])[1]'



from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android.teleguard import logger
from puma.state_graph.puma_driver import PumaDriver, supported_version
from puma.state_graph.action import action
from puma.state_graph.state_graph import StateGraph
from puma.state_graph.state import ContextualState, SimpleState, compose_clicks

APPLICATION_PACKAGE = 'ch.swisscows.messenger.teleguardapp'

CONVERSATION_STATE_TELEGUARD_HEADER = '//android.view.View[@content-desc="TeleGuard"]'
CONVERSATION_STATE_HAMBURGER_MENU = '//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.view.View[1]/android.view.View[2]/android.view.View[3]'
CONVERSATION_STATE_SETTINGS_BUTTON = '//android.widget.ImageView[@content-desc="Settings"]'
CONVERSATION_STATE_ABOUT_BUTTON = '//android.widget.ImageView[@content-desc="About"]'
CONVERSATION_STATE_TELEGUARD_STATUS = '//android.view.View[@content-desc="Online"]|//android.view.View[contains(@content-desc, "Connection to server")]'
CONVERSATION_STATE_ADD_CONTACT = '//android.widget.ImageView[@content-desc="Add contact"]'
CONVERSATION_STATE_EDIT_TEXT = '//android.widget.EditText'
CONVERSATION_STATE_INVITE = '//android.widget.Button[@content-desc="INVITE"]'
CONVERSATION_STATE_YOU_HAVE_BEEN_INVITED = '//android.view.View[contains(@content-desc, "You have been invited")]'
CONVERSATION_STATE_ACCEPT_INVITE = '//android.widget.Button[@content-desc="ACCEPT INVITE"]'

CHAT_STATE_CONVERSATION_NAME = ('//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.view.View[1]/android.view.View[2]/android.widget.ImageView[2][@content-desc]|'
                                '//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.view.View[1]/android.view.View[2]/android.view.View[1][@content-desc]')
CHAT_STATE_TEXT_FIELD = '//android.widget.EditText'
CHAT_STATE_MICROPHONE_BUTTON = '//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.widget.ImageView[4]'
CHAT_STATE_SEND_BUTTON = '//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.widget.ImageView[3]'

SETTINGS_STATE_CHANGE_TELE_GUARD_ID = '//android.view.View[@content-desc="Change TeleGuard ID"]'

ABOUT_STATE_ABOUT = '//android.view.View[@content-desc="About"]'
ABOUT_STATE_TERMS_OF_USE = '//android.view.View[@content-desc=" Terms of use"]'

def go_to_chat(driver: PumaDriver, conversation: str):
    """
    Navigates to a specific chat conversation in the application.

    This function constructs an XPath to locate and click on a conversation element
    based on the conversation name. It is designed to be used within a state transition
    to navigate to a specific chat state.

    :param driver: The PumaDriver instance used to interact with the application.
    :param conversation: The name of the conversation to navigate to.
    """
    xpath = f'//android.widget.ImageView[contains(lower-case(@content-desc), "{conversation.lower()}")] | ' \
            f'//android.view.View[contains(lower-case(@content-desc), "{conversation.lower()}")]'
    driver.driver.find_elements(by=AppiumBy.XPATH, value=xpath)[-1].click()

class TeleGuardChatState(SimpleState, ContextualState):
    """
    A state representing a chat screen in the application.

    This class extends both SimpleState and ContextualState to represent a chat screen
    and validate its context based on the conversation name.
    """

    def __init__(self, parent_state):
        """
        Initializes the ChatState with a parent state.

        :param parent_state: The parent state of this chat state.
        """
        super().__init__(xpaths=[CHAT_STATE_CONVERSATION_NAME, CHAT_STATE_MICROPHONE_BUTTON, CHAT_STATE_TEXT_FIELD],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        """
        Validates the context of the chat state.

        This method checks if the current chat screen matches the expected conversation name.

        :param driver: The PumaDriver instance used to interact with the application.
        :param conversation: The name of the conversation to validate against.
        :return: True if the context is valid, False otherwise.
        """
        if not conversation:
            return True

        content_desc = (driver.get_element(CHAT_STATE_CONVERSATION_NAME).get_attribute('content-desc'))
        return conversation.lower() in content_desc.lower()

@supported_version('4.0.7')
class TeleGuard(StateGraph):
    """
    A class representing a state graph for managing UI states and transitions in the TeleGuard application.

    This class uses a state machine approach to manage transitions between different states
    of the TeleGuard user interface. It provides methods to navigate between states, validate states,
    and handle unexpected states or errors.
    """

    conversations_state = SimpleState( [CONVERSATION_STATE_TELEGUARD_HEADER, CONVERSATION_STATE_HAMBURGER_MENU, CONVERSATION_STATE_TELEGUARD_STATUS], initial_state=True)
    chat_state = TeleGuardChatState(parent_state=conversations_state)
    settings_state = SimpleState([SETTINGS_STATE_CHANGE_TELE_GUARD_ID], parent_state=conversations_state)
    about_screen_state = SimpleState([ABOUT_STATE_ABOUT, ABOUT_STATE_TERMS_OF_USE], parent_state=conversations_state)

    conversations_state.to(chat_state, go_to_chat)
    conversations_state.to(settings_state, compose_clicks([CONVERSATION_STATE_HAMBURGER_MENU, CONVERSATION_STATE_SETTINGS_BUTTON], name='go_to_settings'))
    conversations_state.to(about_screen_state, compose_clicks([CONVERSATION_STATE_HAMBURGER_MENU, CONVERSATION_STATE_ABOUT_BUTTON], name='go_to_about'))

    def __init__(self, device_udid):
        """
        Initializes the TestFsm with a device UDID.

        This class provides an API for interacting with the TeleGuard application.
        It can be used with an emulator or a real device attached to the computer.

        :param device_udid: The unique device identifier for the Android device.
        """
        StateGraph.__init__(self, device_udid, APPLICATION_PACKAGE)

    @action(chat_state)
    def send_message(self, msg: str, conversation: str = None):
        """
        Sends a message in the current chat conversation.

        :param msg: The message to send.
        :param conversation: The name of the conversation to send the message in.
        """
        self.driver.click(CHAT_STATE_TEXT_FIELD)
        self.driver.send_keys(CHAT_STATE_TEXT_FIELD, msg)
        self.driver.click(CHAT_STATE_SEND_BUTTON)

    @action(conversations_state)
    def add_contact(self, teleguard_id: str):
        """
        Adds a contact by TeleGuard ID.

        :param teleguard_id: The TeleGuard ID of the contact to add.
        """
        self.driver.click(CONVERSATION_STATE_HAMBURGER_MENU)
        self.driver.click(CONVERSATION_STATE_ADD_CONTACT)
        self.driver.send_keys(CONVERSATION_STATE_EDIT_TEXT, teleguard_id)
        self.driver.click(CONVERSATION_STATE_INVITE)

    @action(conversations_state)
    def accept_invite(self):
        """
        Accepts an invite from another user.

        If there are multiple invites, only the topmost invite in the UI will be accepted.
        """
        self.driver.swipe_to_click_element(CONVERSATION_STATE_YOU_HAVE_BEEN_INVITED)
        self.driver.click(CONVERSATION_STATE_ACCEPT_INVITE)



from typing import List

from puma.state_graph.puma_driver import PumaDriver
from puma.state_graph.state import compose_clicks


class PopUpHandler:
    """
    Handler for pop-up windows in Android applications.
    """

    def __init__(self, recognize_xpaths: List[str], dismiss_xpaths: List[str]):
        """
        Pop-up handler.

        :param recognize_xpaths: The XPaths to use for recognizing popup windows.
        :param dismiss_xpaths: The XPaths for the element to dismiss the pop-up.
        """
        self.recognize_xpaths = recognize_xpaths
        self.dismiss_xpaths = dismiss_xpaths

    def is_popup_window(self, driver: PumaDriver) -> bool:
        """
        Check if a pop-up is present in the current window

        :param driver: The PumaDriver instance to use for searching the window.
        return: Whether the pop-up window was found or not.
        """
        return all(driver.is_present(xpath) for xpath in self.recognize_xpaths)

    def dismiss_popup(self, driver: PumaDriver):
        """
        Dismiss a pop-up window using the provided xpath.

        :param driver: The PumaDriver instance to use for searching and clicking the button.
        """
        driver.gtl_logger.info('Dismissing pop-up')
        compose_clicks(self.dismiss_xpaths)(driver)


def simple_popup_handler(xpath: str):
    """
    Utility method to create a pop-up handler that uses the same XPath for both recognizing and dismissing the pop-up.

    :param xpath: XPath of the element to click
    :return: PopUpHandler for the provided XPath
    """
    return PopUpHandler([xpath], [xpath])


known_popups = [simple_popup_handler('//android.widget.ImageView[@content-desc="Dismiss update dialog"]'),
                simple_popup_handler(
                    '//android.widget.Button[@resource-id="com.android.permissioncontroller:id/permission_allow_foreground_only_button"]'),
                simple_popup_handler(
                    '//android.widget.Button[@resource-id="com.android.permissioncontroller:id/permission_allow_button"]')]



import os
import re
import unittest
from puma.utils import PROJECT_ROOT
from puma.version import version as setup_version

import sqlite3
import os
import random

def read_release_notes(file_path: str) -> [str]:
    """
    Read the content of the release notes file.
    :param file_path: The path to the release notes file.
    :return: The content of the release notes file.
    """
    with open(file_path, 'r') as file:
        return file.readlines()


class TestVersion(unittest.TestCase):

    def setUp(self):
        self.release_notes_path = f"{PROJECT_ROOT}/RELEASE_NOTES"
        self.release_notes = read_release_notes(self.release_notes_path)
        self.first_version_in_release_notes = self._get_first_version_in_release_notes()

    def _get_first_version_in_release_notes(self):
        first_line = self.release_notes[0]
        match = re.search(r'(\d+\.\d+\.\d+)', first_line)
        return match.group(1) if match else None

    def test_version_in_release_notes_same_as_setup(self):
        self.assertIsNotNone(self.first_version_in_release_notes)
        self.assertEqual(self.first_version_in_release_notes, setup_version,
                         "Version in release notes is not equal to setup version")

    def test_versions_same_as_github(self):
        github_tag_version = os.getenv('GITHUB_TAG_VERSION')
        # Only run this test when a release is made from GitHub (i.e. the above environment variable is set)
        if github_tag_version is None:
            self.skipTest("Skipping GitHub version test as no tag version was passed.")

        self.assertEqual(github_tag_version, self.first_version_in_release_notes,
                         "GitHub tag version is not equal to top version in release notes")
        self.assertEqual(github_tag_version, setup_version,
                         "GitHub tag version is not equal to setup version")

if __name__ == '__main__':

    # Connect to the SQLite database (creates it if it doesn't exist)
    conn = sqlite3.connect('/media/bouke/externe-ssd/dev/libraries/traces/traces-db/src/test/resources/testdata/database/sqlite/blob_in_text_column_benchmark.db')
    cursor = conn.cursor()

    # Insert 10,000 records with random binary data
    for _ in range(10000):
        # Generate random binary data (500-2000 bytes)
        byte_length = random.randint(500,2000)
        random_bytes = os.urandom(byte_length).hex()
        print(random_bytes)
        # cursor.execute('INSERT INTO data (content) VALUES (?)', (random_bytes,))

    # Commit the changes and close the connection
    conn.commit()
    conn.close()

    print("Inserted 10,000 records with random binary data.")



from time import sleep
from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy
from typing_extensions import deprecated

from puma.apps.android.appium_actions import supported_version, AndroidAppiumActions
from puma.apps.android.whatsapp_business.whatsapp_common import WhatsAppCommon


@deprecated('This class does not use the Puma state machine, and will therefore not be maintained. ' +
            'If you want to add functionality, please rewrite this class using StateGraph as the abstract base class.')
@supported_version("2.25.24.78")
class WhatsappBusinessActions(WhatsAppCommon):

    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        """
        Class with an API for WhatsApp for Business Android using Appium. Can be used with an emulator or real device attached to the computer.
        """
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      'com.whatsapp.w4b',
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)

    def change_profile_picture(self, photo_dir_name, index=1):
        self.return_to_homescreen()
        self.open_settings_you()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.ImageView[@content-desc="Edit photo"]').click()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.TextView[@resource-id="android:id/text1" and @text="Add or edit profile photo"]').click()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.TextView[@resource-id="com.whatsapp.w4b:id/row_text" and @text="Gallery"]').click()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.ImageButton[@content-desc="Folders"]').click()
        WhatsAppCommon._find_media_in_folder(self, photo_dir_name, index)
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.Button[@resource-id="com.whatsapp.w4b:id/ok_btn"]').click()

    def set_about(self, about_text: str):
        self.return_to_homescreen()
        self.open_settings_you()
        self.swipe_to_find_element(f'//android.widget.TextView[@text="About"]').click()
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.ImageView[@content-desc="edit"]').click()
        text_box = self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/edit_text")
        text_box.click()
        text_box.clear()
        text_box.send_keys(about_text)
        self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/save_button").click()

    def leave_group(self, group_name):
        self.select_chat(group_name)
        self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/conversation_contact").click()
        self.scroll_to_find_element(text_equals="Exit group").click()
        self.driver.find_element(by=AppiumBy.XPATH, value="//android.widget.Button[@text='Exit group']").click()
        self.return_to_homescreen()

    def send_media(self, directory_name, index=1, caption=None, chat: str = None):
        self._if_chat_go_to_chat(chat)
        # Go to gallery
        self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/input_attach_button").click()
        self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/pickfiletype_gallery_holder").click()

        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.ImageButton[@content-desc="Folders"]').click()
        WhatsAppCommon._find_media_in_folder(self, directory_name, index)
        sleep(0.5)
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value=f'//androidx.compose.ui.platform.ComposeView/android.view.View/android.view.View/android.view.View[5]/android.view.View[3]/android.widget.Button').click()

        if caption:
            sleep(0.5)
            # text_box = self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.TextView[contains(@content-desc, "photos or videos selected")]')
            text_box = self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/caption")
            text_box.send_keys(caption)
            # Clicking the text box after sending keys is required for Whatsapp to notice text has been inserted.
            text_box.click()
            self.driver.back()

        sleep(1)
        self.driver.find_element(by=AppiumBy.ID, value=f"{self.app_package}:id/send").click()



# General
TAB_SWITCH_BUTTON = '//android.widget.ImageButton[@resource-id="com.android.chrome:id/tab_switcher_button"]'
THREE_DOTS = '//android.widget.ImageButton[@resource-id="com.android.chrome:id/menu_button"]'
# Tab overview
SEARCH_TABS = '//android.widget.EditText[@resource-id="com.android.chrome:id/search_box_text" and @text="Search your tabs"]'
SEARCH_INCOGNITO_TABS = '//android.widget.EditText[@resource-id="com.android.chrome:id/search_box_text" and @text="Search your Incognito tabs"]'
STANDARD_TAB_OVERVIEW_BUTTON = '//android.widget.LinearLayout[contains(@content-desc,"standard tab")]'
NEW_TAB_XPATH_TAB_OVERVIEW = '//android.widget.Button[@content-desc="New tab"]'
TAB_LIST = '//*[@resource-id="com.android.chrome:id/tab_list_recycler_view"]'
# Tab
SEARCH_BOX = '//android.widget.EditText[@resource-id="com.android.chrome:id/search_box_text"]'
SEARCH_BOX_ENGINE_ICON = '//android.widget.ImageView[@resource-id="com.android.chrome:id/search_box_engine_icon"]'
URL_BAR = '//android.widget.EditText[@resource-id="com.android.chrome:id/url_bar"]'
NEW_TAB_FROM_CURRENT_TAB = '//android.widget.ImageButton[@content-desc="New tab"]'
# Menu
NEW_TAB_INCOGNITO_TITLE = '//android.widget.TextView[@resource-id="com.android.chrome:id/new_tab_incognito_title"]'
BOOKMARK_THIS_PAGE_BUTTON = '//*[lower-case(@content-desc)="bookmark"]'
EDIT_BOOKMARK_BUTTON = '//*[lower-case(@content-desc)="edit bookmark"]'
NEW_INCOGNITO_TAB_BUTTON = '//android.widget.TextView[@resource-id="com.android.chrome:id/title" and @text="New Incognito tab"]'
OPEN_BOOKMARKS = '//android.widget.TextView[@resource-id="com.android.chrome:id/menu_item_text" and @text="Bookmarks"]'
# Bookmarks
MOBILE_BOOKMARKS = '//android.widget.TextView[@resource-id="com.android.chrome:id/title" and @text="Mobile bookmarks"]'
DELETE_BOOKMARK = '//android.widget.Button[lower-case(@content-desc)="delete bookmarks"]'
CLOSE_BOOKMARKS = '//android.widget.Button[@resource-id="com.android.chrome:id/close_menu_id"]'
FIRST_BOOKMARK = '//android.widget.LinearLayout[@resource-id="com.android.chrome:id/container"]'
BOOKMARKS_PAGE_TITLE = '//android.widget.TextView[@text="Bookmarks"]'
BOOKMARKS_SORT_VIEW = '//android.widget.Button[@content-desc="Sort and view options"]'
BOOKMARKS_CREATE_FOLDER = '//android.widget.Button[@content-desc="Create new folder"]'
BOOKMARKS_GO_BACK = '//android.widget.ImageButton[@content-desc="Go back"]'
# Popups
NEW_ADS_PRIVACY_TEXT_XPATH = '//android.widget.TextView[@text="New ads privacy feature now available"]'
OTHER_ADS_PRIVACY_TEXT_XPATH = '//android.widget.TextView[@text="Other ad privacy features now available"]'
TURN_ON_AD_PRIVACY_TEXT_XPATH = '//android.widget.TextView[@text="Turn on an ad privacy feature"]'
CHROME_NOTIFICATIONS_TEXT_XPATH = '//android.widget.TextView[@text="Chrome notifications make things easier"]'
ACK_BUTTON_XPATH = '//android.widget.Button[@resource-id="com.android.chrome:id/ack_button"]'
MORE_BUTTON_XPATH = '//android.widget.Button[@resource-id="com.android.chrome:id/more_button"]'
NEGATIVE_BUTTON_XPATH = '//android.widget.Button[@resource-id="com.android.chrome:id/negative_button"]'


import logging
from pathlib import Path
from sys import stdout

from puma.utils import LOG_FOLDER, PUMA_INIT_TIMESTAMP


def create_gtl_logger(udid: str) -> logging.Logger:
    """
    Create a Puma Ground Truth Logger, specific for one device.
    This logger will log the ground truth of actions taken on this device, including navigation and UI interactions.
    Each log line will include the time and device udid so a clear timeline of events can be tracked on each device.

    :param udid: the device id of the device this logger tracks.
    """
    gtl_logger = logging.getLogger(f'{udid}')

    # format of log lines
    formatter = logging.Formatter(fmt=f'%(asctime)s [%(levelname)s] [{udid}] %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    # file name of log file
    file_handler = logging.FileHandler(Path(LOG_FOLDER) / f'{PUMA_INIT_TIMESTAMP}_gtl.log')
    file_handler.setFormatter(formatter)

    gtl_logger.addHandler(file_handler)

    return gtl_logger



from puma.apps.android.snapchat.xpaths import *
from puma.state_graph.action import action
from puma.state_graph.popup_handler import PopUpHandler
from puma.state_graph.puma_driver import PumaDriver, supported_version
from puma.state_graph.state import SimpleState, ContextualState, compose_clicks
from puma.state_graph.state_graph import StateGraph

APPLICATION_PACKAGE = 'com.snapchat.android'


def go_to_chat(driver: PumaDriver, conversation: str):
    """
    Navigates to a specific chat conversation in the application.

    This function constructs an XPath to locate and click on a conversation element
    based on the conversation name. It is designed to be used within a state transition
    to navigate to a specific chat state.

    :param driver: The PumaDriver instance used to interact with the application.
    :param conversation: The name of the conversation to navigate to.
    """

    driver.click(CHAT_CONVERSATION.format(conversation=conversation))


def go_to_snap(driver: PumaDriver, caption: str = None):
    """
    Navigates to the snap state.

    This function optionally sets a caption on the photo that was just taken.

    :param driver: The PumaDriver instance used to interact with the application.
    :param caption: The caption to add on the photo.
    """
    if caption:
        Snapchat.add_caption(driver, caption)
    driver.click(SENT_TO)


class SnapchatChatState(SimpleState, ContextualState):
    """
    A state representing a chat screen in the application.

    This class extends both SimpleState and ContextualState to represent a chat screen
    and validate its context based on the conversation name.
    """

    def __init__(self, parent_state):
        """
        Initializes the ChatState with a parent state.

        :param parent_state: The parent state of this chat state.
        """
        super().__init__(xpaths=[CONVERSATION_TITLE, CHAT_INPUT],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        """
        Validates the context of the chat state.

        This method checks if the current chat screen matches the expected conversation name.

        :param driver: The PumaDriver instance used to interact with the application.
        :param conversation: The name of the conversation to validate against.
        :return: True if the context is valid, False otherwise.
        """
        if not conversation:
            return True

        return driver.is_present(CONVERSATION_TITLE_TEXT.format(conversation=conversation.lower()))


@supported_version('12.89.0.40')
class Snapchat(StateGraph):
    """
    A class representing a state graph for managing UI states and transitions in the Snapchat application.

    This class uses a state machine approach to manage transitions between different states
    of the Snapchat user interface. It provides methods to navigate between states, validate states,
    and handle unexpected states or errors.
    """
    camera_state = SimpleState([CAMERA_PAGE], initial_state=True)
    conversation_state = SimpleState([FEED_NEW_CHAT], parent_state=camera_state)
    chat_state = SnapchatChatState(parent_state=conversation_state)
    captured_state = SimpleState([SENT_TO], parent_state=camera_state, invalid_xpaths=[ALERT_DIALOG_DESCRIPTION])
    snap_state = SimpleState([NEW_STORY], parent_state=captured_state)

    camera_state.to(conversation_state, compose_clicks([CHAT_TAB], name='press_chat_tab'))
    conversation_state.to(chat_state, go_to_chat)
    camera_state.to(captured_state, compose_clicks([CAMERA_CAPTURE], name='press_camera_capture'))
    captured_state.to(snap_state, go_to_snap)

    def __init__(self, device_udid):
        """
        Initializes Snapchat with a device UDID.

        This class provides an API for interacting with the Snapchat application.
        It can be used with an emulator or a real device attached to the computer.

        :param device_udid: The unique device identifier for the Android device.
        """
        StateGraph.__init__(self, device_udid, APPLICATION_PACKAGE)
        self.add_popup_handler(PopUpHandler([ALERT_DIALOG_DESCRIPTION], [DISCARD_BUTTON]))

    @action(chat_state)
    def send_message(self, message: str, conversation: str = None):
        """
        Sends a message in the specified chat conversation.

        :param message: The message to send.
        :param conversation: The name of the conversation to send the message in. If nothing is specified, the current chat will be used.
        """
        self.driver.click(CHAT_INPUT)
        self.driver.send_keys(CHAT_INPUT, message)
        self.driver.press_enter()

    @action(camera_state)
    def toggle_camera(self):
        """
        Toggles camera.
        Default state is front facing camera. After closing the app the camera is faced the same direction as it was when previously closed.
        """
        self.driver.click(TOGGLE_CAMERA)

    @staticmethod
    def add_caption(driver: PumaDriver, caption: str):
        driver.click(FULL_SCREEN_SURFACE_VIEW)
        caption_field = driver.get_element(CAPTION_EDIT)
        caption_field.send_keys(caption)
        driver.click(FULL_SCREEN_SURFACE_VIEW)

    @action(snap_state, end_state=conversation_state)
    def send_snap_to(self, recipients: list[str], caption: str = None):
        """
        Sends a snap to recipients.

        :param recipients: The recipients to send the snap to.
        """
        for recipient in recipients:
            self.driver.click(RECIPIENTS_TO_ADD.format(recipient=recipient))
        self.driver.click(SEND_BUTTON)

    @action(snap_state, end_state=camera_state)
    def send_snap_to_my_story(self, caption: str = None):
        """
        Sends a snap to my story.
        """
        self.driver.click(MY_STORY)
        self.driver.click(SEND_BUTTON)



from time import sleep
from typing import Dict

from puma.state_graph import logger
from puma.state_graph.popup_handler import known_popups, PopUpHandler
from puma.state_graph.puma_driver import PumaDriver, PumaClickException
from puma.state_graph.state import State, ContextualState, Transition, _shortest_path
from puma.state_graph.utils import safe_func_call, filter_arguments, is_valid_package_name


class StateGraphMeta(type):
    """
    Metaclass for creating and validating StateGraph classes.

    This metaclass is responsible for collecting states and transitions defined within a class,
    and performing validation to ensure the integrity of the state graph. It ensures that there
    is exactly one initial state, that all states are reachable from the initial state, and
    that there are no duplicate state names or transitions.
    """

    def __new__(cls, name, bases, namespace):
        """
        Creates a new class using the StateGraphMeta metaclass.

        This method collects states and transitions from the class namespace and performs
        validation on the state graph.

        :param name: The name of the class being created.
        :param bases: The base classes of the class being created.
        :param namespace: The namespace dictionary containing the class attributes.
        :return: The newly created class.
        """
        new_class = super().__new__(cls, name, bases, namespace)

        # Skip validation for the base PumaUIGraph class
        if name == 'StateGraph':
            return new_class

        # collect states and transitions
        states: list[State] = []
        transitions = []
        for key, value in namespace.items():
            if isinstance(value, State):
                states.append(value)
                value.id = key
            if isinstance(value, Transition):
                transitions.append(value)
        new_class.states = states
        new_class.transitions = [transition for state in states for transition in state.transitions]
        new_class.initial_state = next(s for s in states if s.initial_state)

        # validation
        StateGraphMeta._validate_graph(states)

        return new_class

    @staticmethod
    def _validate_graph(states: list[State]):
        """
        Validates the state graph to ensure it meets specific criteria.

        This method checks for the presence of exactly one initial state, ensures that
        the initial state is not a ContextualState, verifies that all ContextualStates
        have a parent state, and checks that all states are reachable from the initial
        state and vice versa. It also ensures there are no duplicate state names or
        transitions.

        :param states: A list of states in the state graph.
        :raises ValueError: If any of the validation checks fail.
        """
        # There can only be one initial state
        StateGraphMeta._validate_only_one_initial_state(states)
        # Initial state cannot be Contextual state
        StateGraphMeta._validate_initial_state(states)
        # You need to be able to go from the initial state to each other state and back
        StateGraphMeta._validate_every_state_reachable(states)
        # Contextual states need parent state
        StateGraphMeta._validate_contextual_states(states)
        # No duplicate ids for states
        StateGraphMeta._validate_no_duplicate_ids(states)
        # No duplicate transitions: max one transition between each 2 states
        StateGraphMeta._validate_no_duplicate_transitions(states)

    @staticmethod
    def _validate_only_one_initial_state(states):
        initial_states = [s for s in states if s.initial_state]
        if len(initial_states) == 0:
            raise ValueError(f'Graph needs an initial state')
        elif len(initial_states) > 1:
            raise ValueError(f'Graph can only have 1 initial state, currently more defined: {initial_states}')
        initial_state = initial_states[0]
        return initial_state

    @staticmethod
    def _validate_initial_state(states):
        initial_state = next(s for s in states if s.initial_state)
        if isinstance(initial_state, ContextualState):
            raise ValueError(f'Initial state ({initial_state}) Cannot be an Contextual state')

    @staticmethod
    def _validate_every_state_reachable(states):
        initial_state = next(s for s in states if s.initial_state)
        unreachable_transitions = [s for s in states if not s.initial_state and not (
                    bool(_shortest_path(initial_state, s)))]
        unreachable_parents = [s for s in states if not s.initial_state and not (
            bool(_shortest_path(s, initial_state)))]
        if unreachable_parents or unreachable_parents:
            raise ValueError(
                f"Some states cannot be reached from the initial state '{initial_state}'.\n"
                f"Add the missing transitions or missing parent states to your Puma StateGraph to fix this issue.\n"
                f"Missing transitions: {unreachable_transitions}\n"
                f"Missing parent states: {unreachable_parents}"
            )

    @staticmethod
    def _validate_contextual_states(states):
        contextual_state_without_parent = [s for s in states if
                                           isinstance(s, ContextualState) and s.parent_state is None]
        if contextual_state_without_parent:
            raise ValueError(f'Contextual states without parent are not allowed: {contextual_state_without_parent}')

    @staticmethod
    def _validate_no_duplicate_ids(states):
        seen = set()
        duplicate_ids = {s.id for s in states if s.id in seen or seen.add(s.id)}
        if duplicate_ids:
            raise ValueError(f"States must have a unique id. Multiple sates have the id {duplicate_ids}")

    @staticmethod
    def _validate_no_duplicate_transitions(states):
        for s in states:
            seen = set()
            duplicates = {t.to_state for t in s.transitions if t.to_state in seen or seen.add(t.to_state)}
            if duplicates:
                raise ValueError(
                    f"State {s} has invalid transitions: multiple transitions defined to neighboring state(s) {duplicates}")


class StateGraph(metaclass=StateGraphMeta):
    """
    A class representing a state graph for managing UI states and transitions in an application.

    This class uses a state machine approach to manage transitions between different states
    of a user interface. It initializes with a device and application package, and provides
    methods to navigate between states, validate states, and handle unexpected states or errors.
    """

    def __init__(self, device_udid: str, app_package: str, appium_server: str = 'http://localhost:4723', desired_capabilities: Dict[str, str] = None):
        """
        Initializes the StateGraph with a device and application package.

        :param device_udid: The unique device identifier.
        :param app_package: The package name of the application.
        :param desired_capabilities: desired capabilities as passed to the Appium webdriver.
        """
        if not is_valid_package_name(app_package):
            raise ValueError(f'The provided package name is invalid: {app_package}')
        self.current_state = self.initial_state
        self.driver = PumaDriver(device_udid, app_package, appium_server=appium_server, desired_capabilities=desired_capabilities)
        self.app_popups = []
        self.try_restart = True
        self.gtl_logger = self.driver.gtl_logger

    def go_to_state(self, to_state: State | str, **kwargs) -> bool:
        """
        Navigates to a specified state from the current state.

        :param to_state: The destination state or destination state name.
        :param kwargs: Additional keyword arguments to pass to state validation and transition functions.
        :return: True if the transition to the desired state is successful.
        """
        max_transitions = len(self.states) * 2 + 5
        counter = 0
        if to_state not in self.states:
            raise ValueError(f"{to_state.id} is not a known state in this PumaUiGraph")
        kwargs['driver'] = self.driver
        try:
            self._validate_state(self.current_state, **kwargs)
        except PumaClickException as pce:
            logger.warning(f"Initial state validation encountered a problem {pce}")
        while self.current_state != to_state and counter < max_transitions:
            counter += 1
            try:
                transition = self._find_shortest_path(to_state)[0]

                self.gtl_logger.info(f"Going from state '{self.current_state}' to '{to_state}', calling transition '{transition.ui_actions.__name__}'")
                safe_func_call(transition.ui_actions, **kwargs)

                self._validate_state(transition.to_state, **kwargs)
            except PumaClickException as pce:
                logger.warning(f"Transition or state validation failed, recover? {pce}")
        if counter >= max_transitions:
            raise ValueError(f"Too many transitions, state is unrecoverable")
        return True

    def _validate_state(self, expected_state: State, **kwargs):
        """
        Validates the current state against an expected state.

        This method checks if the current state matches the expected state and handles
        any discrepancies by attempting to recover the expected state or context.

        :param expected_state: The state that is expected to be the current state.
        :param kwargs: Additional keyword arguments for context validation.
        """
        # validate
        valid = expected_state.validate(self.driver)
        context_valid = True
        if isinstance(expected_state, ContextualState):
            context_valid = safe_func_call(expected_state.validate_context, **kwargs)

        # handle validation results
        if not valid:
            self.gtl_logger.error(f"Expected to be in state '{expected_state}', but the validation failed")

             # state is totally unexpected
            self.recover_state(expected_state)
            self._validate_state(self.current_state, **kwargs)
        elif not context_valid:
            relevant_kwargs = filter_arguments(expected_state.validate_context, **kwargs)
            self.gtl_logger.error(f"Was in the expected state '{expected_state}', but context '{relevant_kwargs}' did not match")
            # correct state, but wrong context (e.g. we want a conversation with Alice, but we're in a conversation with Bob)
            # recovery: always go back to the parent state
            # Go to the first non-contextual parent state, context is lost when recovering from a contextual state
            go_to_state = expected_state.parent_state
            while isinstance(go_to_state, ContextualState):
                go_to_state = go_to_state.parent_state
            self.go_to_state(go_to_state)
        else:
            self.gtl_logger.info(f"Validated that current state is the expected state '{expected_state}'")
            self.current_state = expected_state

    def recover_state(self, expected_state):
        """
        Attempts to recover the expected state if the current state is not the expected state.

        This method ensures the application is active, handles pop-ups, and attempts to
        identify and recover the current state if it does not match the expected state.

        :param expected_state: The state that is expected to be the current state.
        """
        current_state = self.current_state if self.current_state != expected_state else 'unknown'
        self.gtl_logger.info(f"Recovering state to '{expected_state}', current state is '{current_state}'")
        # Ensure app active
        if not self.driver.app_open():
            self.driver.activate_app()

        if self.current_state.validate(self.driver):
            return

        # pop-ups
        self._handle_popups()

        if self.current_state.validate(self.driver):
            return

        # Search state
        self._search_state(expected_state)

    def _handle_popups(self):
        clicked = True
        while clicked:
            clicked = False
            for popup_handler in known_popups + self.app_popups:
                if popup_handler.is_popup_window(self.driver):
                    popup_handler.dismiss_popup(self.driver)
                    clicked = True

    def _search_state(self, expected_state: State):
        current_states = [s for s in self.states if s.validate(self.driver)]
        if len(current_states) != 1:
            if not self.try_restart:
                if len(current_states) > 1:
                    raise ValueError(f"More than one state matches the current UI. Write stricter XPaths. states: {current_states}")
                else:
                    raise ValueError("Unknown state, cannot recover")
            logger.warning(f'Not in a known state. Restarting app {self.driver.app_package} once')
            self.driver.restart_app()
            sleep(3)
            self.try_restart = False
            return
        self.gtl_logger.info(
            f"Was in unknown state, expected '{expected_state}'. Recognized state, setting current state to '{current_states[0]}'")
        self.current_state = current_states[0]

    def add_popup_handler(self, popup_handler: PopUpHandler):
        """
        Adds a pop-up handler to manage application pop-ups.

        :param popup_handler: The pop-up handler to be added.
        """
        self.app_popups.append(popup_handler)

    def add_popup_handlers(self, *popup_handlers: PopUpHandler):
        """
        Adds multiple pop-up handlers to manage application pop-ups.

        :param popup_handlers: The pop-up handlers to be added.
        """
        self.app_popups.extend(popup_handlers)

    def _find_shortest_path(self, destination: State | str) -> list[Transition] | None:
        """
        Finds the shortest path in terms of transitions to the desired state.

        :param destination: The destination state or state name.
        :return: A list of transitions representing the shortest path to the destination state, or None if no path is found.
        """
        return _shortest_path(self.current_state, destination)



import io

from PIL import Image
from selenium.webdriver.remote.webelement import WebElement


def show_elements_as_picture_popups(elements: WebElement | list[WebElement]):
    """
    Generate a pup-up image for each element, showing exactly and only the UI part of
    given element. Useful for debugging, for example when you select a set of elements
    using XPATH and you want to see which ones matched.

    :param elements: elements to show
    """
    elements = elements if isinstance(elements, list) else [elements]
    for element in elements:
        image = Image.open(io.BytesIO(element.screenshot_as_png))
        image.show(title=element.id)



import inspect
from typing import Callable, OrderedDict, Any

from puma.state_graph.puma_driver import PumaClickException
from puma.state_graph.state import State
from puma.state_graph.state_graph import StateGraph
from puma.state_graph.utils import filter_arguments


def _assert_verify_with_function_is_valid(verify_with: type[Any]):
    """
    Validate that the 'verify_with' passed to the action is valid, else throw an error.

    :param verify_with: the object passed as 'verify_with'
    :raises TypeError: if the verification function is not a callable
    """
    if not isinstance(verify_with, Callable):
        raise TypeError(f"'verify_with' must be a callable, instead is: {type(verify_with)}")


def _execute_post_action_verification(puma_ui_graph: StateGraph, action: Callable, verify_with: Callable, arguments: OrderedDict[str, Any]):
    """
    Run the given post action verification function and log the results. Returns to the current state
    of the graph at the end of the verification execution.

    :param puma_ui_graph: the application state graph, in the state after the execution of the action
    :param action: function executed as action
    :param verify_with: the post action verification function
    :param arguments: the bound arguments of the action which will be passed to the verification
    """
    gtl_logger = puma_ui_graph.gtl_logger
    # store current state, so we can return to it at the end
    state_before_verify_with = puma_ui_graph.current_state

    # run the post action if present, catching any domain exceptions,
    # since these should not interrupt the rest of the actions
    bound_args = filter_arguments(verify_with, app=puma_ui_graph, **arguments)
    try:
        success = verify_with(**bound_args.arguments)
    except PumaClickException as e:
        gtl_logger.warn(f"Verifying with '{verify_with.__name__}' failed due to exception: {str(e)}")
    else:
        # enforce type here, since we have no typed API call to enforce it with
        if not isinstance(success, bool):
            raise ValueError(f"result of 'verify_with' should be a bool, instead is: {type(success)}, with value: {success}")

        gtl_logger.info(f"Action '{action.__name__}' {'succeeded' if success else 'failed'}")

    # always return to our original state, even if verification failed
    puma_ui_graph.go_to_state(state_before_verify_with, **arguments)


def action(state: State, end_state: State = None):
    """
    Decorator to wrap a function with logic to ensure a specific state before execution.

    This decorator ensures that the application is in the specified state before executing
    the wrapped function. It is useful for performing actions within an app, such as sending
    a message, while ensuring the correct state. If a PumaClickException occurs during the
    execution of the function, it attempts to recover the state and retry the function execution.

    An action can be verified by passing a function as a named argument with name 'verify_with'
    to the function which is decorated. If you pass a static method as the 'verify_with' function,
    it should have a parameter named 'app' in which the application instance will be injected.
    You can instead also pass a reference to a function of an application instance (i.e. a bound method),
    in which case 'self' will simply refer to the application itself.

    The 'verify_with' function will be called at the end of the action's execution. Inside the verification
    function you can change to other states, the framework will ensure it returns to the state it was
    at the end of the action. The return value of the 'verify_with' function must be a bool signaling
    if the action was executed correctly. This will be logged by the framework.

    :param state: The target state to ensure before executing the decorated function.
    :param end_state: Defines if this action ends in a different state (Optional)
    :return: A decorator function that wraps the provided function with state assurance logic.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            """
            Wrapper function that ensures the correct state and handles exception recovery.

            :param args: Positional arguments to pass to the decorated function.
            :param kwargs: Keyword arguments to pass to the decorated function.
            :return: The result of the decorated function.
            """
            verify_with = None
            if 'verify_with' in kwargs.keys():
                verify_with = kwargs.pop('verify_with')
                _assert_verify_with_function_is_valid(verify_with)

            # check that our decorated function has no parameter verify_with, else it would be passed automatically
            action_signature = inspect.signature(func)
            if 'verify_with' in action_signature.parameters:
                raise ValueError(f"the action '{func.__name__}' can't contain a parameter named 'verify_with'")

            bound_args = action_signature.bind(*args, **kwargs)
            bound_args.apply_defaults()
            arguments = bound_args.arguments
            arguments.pop('self')
            puma_ui_graph = args[0]
            # get the ground truth logger to log these actions
            gtl_logger = puma_ui_graph.gtl_logger
            try:
                puma_ui_graph.try_restart = True
                puma_ui_graph.go_to_state(state, **arguments)
                try:
                    gtl_logger.info(
                        f"Executing action '{func.__name__}' with arguments: {args[1:]} and keyword arguments: {kwargs} for application: {puma_ui_graph.__class__.__name__}")
                    result = func(*args, **kwargs)
                except:
                    gtl_logger.info(f"Failed to execute action '{func.__name__}'.")
                    puma_ui_graph.recover_state(state)
                    puma_ui_graph.go_to_state(state, **arguments)
                    gtl_logger.info(f"Retrying action '{func.__name__}'")
                    result = func(*args, **kwargs)

                gtl_logger.info(
                        f"Executed action '{func.__name__}' with arguments: {args[1:]} and keyword arguments: {kwargs} for application: {puma_ui_graph.__class__.__name__}")

                if verify_with is not None:
                    gtl_logger.info(f"Verifying action with '{verify_with.__name__}' using arguments: {args[1:]} and keyword arguments: {kwargs} for application: {puma_ui_graph.__class__.__name__}")
                    _execute_post_action_verification(puma_ui_graph, func, verify_with, arguments)

                puma_ui_graph.try_restart = True

                if end_state:
                    puma_ui_graph.current_state = end_state
                return result
            except Exception as e:
                gtl_logger.error("Unexpected exception while executing an action", e)
                raise e


        return wrapper

    return decorator



def build_resource_id_xpath_widget(widget_type: str, package_name: str, resource_id: str) -> str:
    return _build_resource_id_xpath(f'android.widget.{widget_type}', package_name, resource_id)

def build_resource_id_xpath(package_name: str, resource_id: str) -> str:
    return _build_resource_id_xpath('*', package_name, resource_id)

def _build_resource_id_xpath(class_name: str, package_name: str, resource_id: str) -> str:
    return f'//{class_name}[@resource-id="{package_name}:id/{resource_id}"]'

def build_content_desc_xpath_widget(widget_type: str, content_desc: str) -> str:
    return _build_content_desc_xpath(f'android.widget.{widget_type}', content_desc)

def build_content_desc_xpath(content_desc: str) -> str:
    return _build_content_desc_xpath('*', content_desc)

def _build_content_desc_xpath(class_name: str, content_desc: str) -> str:
    return f'//{class_name}[@content-desc="{content_desc}"]'

def build_text_xpath_widget(widget_type: str, text: str) -> str:
    return _build_text_xpath(f'android.widget.{widget_type}', text)

def build_text_xpath(text: str) -> str:
    return _build_text_xpath('*', text)

def _build_text_xpath(class_name: str, text: str) -> str:
    return f'//{class_name}[@text="{text}"]'

def build_resource_id_text_xpath_widget(widget_type: str, package_name: str, resource_id: str, text: str) -> str:
    return _build_resource_id_text_xpath(f'android.widget.{widget_type}', package_name, resource_id, text)

def build_resource_id_text_xpath(package_name: str, resource_id: str, text: str) -> str:
    return _build_resource_id_text_xpath('*', package_name, resource_id, text)

def _build_resource_id_text_xpath(class_name: str, package_name: str, resource_id: str, text: str) -> str:
    return f'//{class_name}[@resource-id="{package_name}:id/{resource_id}" and @text="{text}"]'


from time import sleep
from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy
from typing_extensions import deprecated

from puma.apps.android import log_action
from puma.apps.android.appium_actions import AndroidAppiumActions, supported_version
from puma.apps.android.open_camera import logger
from puma.state_graph.action import action
from puma.state_graph.popup_handler import PopUpHandler
from puma.state_graph.state import State, SimpleState, compose_clicks
from puma.state_graph.state_graph import StateGraph

# Take photo/video screen
OPEN_CAMERA_PACKAGE = 'net.sourceforge.opencamera'
TAKE_PHOTO = '//android.widget.ImageButton[@resource-id="net.sourceforge.opencamera:id/take_photo" and @content-desc="Take Photo"]'
TAKE_VIDEO = '//android.widget.ImageButton[@resource-id="net.sourceforge.opencamera:id/take_photo" and @content-desc="Start recording video"]'
# Shutter button resource id is the same for photo and video
SHUTTER_BUTTON = '//android.widget.ImageButton[@resource-id="net.sourceforge.opencamera:id/take_photo"]'
PHOTO_MODE_XPATH = '//android.widget.ImageButton[@content-desc="Switch to photo mode"]'
VIDEO_MODE_XPATH = '//android.widget.ImageButton[@content-desc="Switch to video mode"]'
SWITCH_CAMERA_XPATH = '//android.widget.ImageButton[@resource-id="net.sourceforge.opencamera:id/switch_camera"]'
ZOOM_SEEKBAR_XPATH = '//android.widget.SeekBar[@content-desc="Zoom"]'

# Popups
POPUP_TITLE = '//android.widget.TextView[@resource-id="android:id/alertTitle" and @text="Open Camera"]'
POPUP_OK_BUTTON = '//android.widget.Button[@resource-id="android:id/button1" and @text="OK"]'


@supported_version("1.55")
class OpenCamera(StateGraph):
    # States
    take_photo_state = SimpleState(xpaths=[TAKE_PHOTO, VIDEO_MODE_XPATH], initial_state=True)
    take_video_state = SimpleState(xpaths=[TAKE_VIDEO, PHOTO_MODE_XPATH])

    # Transitions
    take_photo_state.to(take_video_state, compose_clicks([VIDEO_MODE_XPATH], name='go_to_video_mode'))
    take_video_state.to(take_photo_state, compose_clicks([PHOTO_MODE_XPATH], name='go_to_photo_mode'))

    def __init__(self, device_udid):
        StateGraph.__init__(self, device_udid, OPEN_CAMERA_PACKAGE)
        self.add_popup_handler(PopUpHandler([POPUP_TITLE], [POPUP_OK_BUTTON]))

    @action(take_photo_state)
    def take_picture(self, front_camera: bool = False, zoom_amount: float = 0.0):
        """
        Takes a picture using the camera.
        :param front_camera: Default False. If True, uses the front-facing camera; otherwise, uses the rear camera.
        :param zoom_amount: The zoom amount between 0.0 and 1.0. This is not the true zoom level on the phone, but the
        fraction of the scroll bar.
        """
        self._ensure_correct_view(front_camera)
        self._zoom(zoom_amount)
        self._click_shutter()

    @action(take_video_state)
    def take_video(self, duration: int, front_camera: bool = False, zoom_amount: float = 0.0):
        """
        Takes a video of a given duration.
        :param duration: The duration of the video in seconds.
        :param front_camera: Whether the front or back camera should be used
        :param zoom_amount: The zoom level between 0.0 and 1.0
        """
        self._ensure_correct_view(front_camera)
        self._zoom(zoom_amount)
        self._click_shutter()
        try:
            sleep(duration)
        finally:
            self._click_shutter()

    def _zoom(self, zoom_amount: float):
        """
        Zooms to a given level between 0.0 (zoomed out completely) and 1.0 (zoomed in completely).
        :param zoom_amount: The zoom level between 0.0 and 1.0. This is not the true zoom level on the phone, but the
        fraction of the scroll bar.
        """
        if zoom_amount < 0 or zoom_amount > 1:
            raise ValueError(f'Invalid zoom level: {zoom_amount}. Zoom level needs to be between 0.0 and 1.0')
        zoom_slider = self.driver.get_element(ZOOM_SEEKBAR_XPATH)
        x = zoom_slider.location.get('x')
        y = zoom_slider.location.get('y')
        width = zoom_slider.size.get('width')
        height = zoom_slider.size.get('height')
        # We need to click slightly inside the bounds of the slider, so we don't use the full height
        height = int(0.95 * height)
        x_to_tap = x + (width // 2)
        y_to_tap = y + (height * (1 - zoom_amount))
        self.driver.tap((x_to_tap, y_to_tap))

    def _click_shutter(self):
        """
        Presses the shutter to take a picture or start/stop recording.
        """
        self.driver.click(SHUTTER_BUTTON)

    def _ensure_correct_view(self, front_camera: bool):
        """
        Switches between the front and back camera.
        """
        switch_camera_button = self.driver.get_element(SWITCH_CAMERA_XPATH)
        currently_in_front = 'back' in switch_camera_button.get_attribute("content-desc")
        if currently_in_front != front_camera:
            switch_camera_button.click()
            self.gtl_logger.info("Switched camera view")



from time import sleep

from puma.apps.android.appium_actions import supported_version
from puma.apps.android.google_chrome import logger
from puma.apps.android.google_chrome.states import BookmarksFolder, CurrentTab
from puma.apps.android.google_chrome.xpaths import *
from puma.state_graph.action import action
from puma.state_graph.popup_handler import PopUpHandler
from puma.state_graph.state import SimpleState, compose_clicks
from puma.state_graph.state_graph import StateGraph

GOOGLE_CHROME_PACKAGE = 'com.android.chrome'


@supported_version("141.0.7390.111")
class GoogleChrome(StateGraph):
    """
    A class representing the Google Chrome application on Android devices.

    The Google Chrome browser application does not really fit the parent-child relationship of states well, as most
    states can be reached from most other states. However, to keep to the structure of the StateGraph, we have chosen
    the tab overview as the parent state of most states, as it is the most neutral state from which other states can be
    reached. Moreover, for simplicity, we have not implemented all possible transitions between states.
    """
    # States
    tab_overview_state = SimpleState(xpaths=[TAB_LIST, SEARCH_TABS])
    incognito_tab_overview_state = SimpleState(xpaths=[TAB_LIST, SEARCH_INCOGNITO_TABS],
                                               parent_state=tab_overview_state,
                                               parent_state_transition=compose_clicks([STANDARD_TAB_OVERVIEW_BUTTON], 'go_to_tab_overview'))
    new_tab_state = SimpleState(xpaths=[SEARCH_BOX, SEARCH_BOX_ENGINE_ICON],
                                initial_state=True,
                                parent_state=tab_overview_state,
                                parent_state_transition=compose_clicks([TAB_SWITCH_BUTTON],'go_to_tab_overview'))
    current_tab_state = CurrentTab(parent_state=tab_overview_state)
    new_incognito_tab_state = SimpleState(xpaths=[URL_BAR, NEW_TAB_INCOGNITO_TITLE],
                                          parent_state=incognito_tab_overview_state,
                                          parent_state_transition=compose_clicks([TAB_SWITCH_BUTTON], 'go_to_incognito_tab_overview'))
    bookmarks_state = SimpleState(xpaths=[BOOKMARKS_SORT_VIEW, BOOKMARKS_CREATE_FOLDER, BOOKMARKS_PAGE_TITLE],
                                  parent_state=current_tab_state,
                                  parent_state_transition=compose_clicks([CLOSE_BOOKMARKS], 'go_to_current_tab'))
    bookmarks_folder_state = BookmarksFolder(parent_state=bookmarks_state)
    # Transitions
    tab_overview_state.to(new_tab_state, compose_clicks([NEW_TAB_XPATH_TAB_OVERVIEW], 'go_to_tab_overview'))
    tab_overview_state.to(current_tab_state, current_tab_state.switch_to_tab)
    tab_overview_state.to(new_incognito_tab_state,
                          compose_clicks([THREE_DOTS, NEW_INCOGNITO_TAB_BUTTON], 'go_to_incognito_state'))
    current_tab_state.to(new_tab_state, compose_clicks([NEW_TAB_FROM_CURRENT_TAB], 'go_to_current_tab'))
    current_tab_state.to(bookmarks_state, compose_clicks([THREE_DOTS, OPEN_BOOKMARKS], 'go_to_current_tab'))
    bookmarks_state.to(bookmarks_folder_state, bookmarks_folder_state.go_to_bookmarks_folder)

    def __init__(self, device_udid):
        """
        Initializes Google Chrome with a device UDID.

        :param device_udid: The unique device identifier for the Android device.
        """
        StateGraph.__init__(self, device_udid, GOOGLE_CHROME_PACKAGE)
        self.add_popup_handlers(PopUpHandler([NEW_ADS_PRIVACY_TEXT_XPATH], [ACK_BUTTON_XPATH]),
                                PopUpHandler([OTHER_ADS_PRIVACY_TEXT_XPATH], [MORE_BUTTON_XPATH, ACK_BUTTON_XPATH]),
                                PopUpHandler([TURN_ON_AD_PRIVACY_TEXT_XPATH], [ACK_BUTTON_XPATH]),
                                PopUpHandler([CHROME_NOTIFICATIONS_TEXT_XPATH], [NEGATIVE_BUTTON_XPATH]),
                                )

    @action(current_tab_state)
    def visit_url(self, url_string: str, tab_index: int):
        """
        Visits a url in an existing tab.
        Note that if you supply a tab index that is a new tab, the action will fail. Use go_to_new_tab instead.
        :param url_string: The argument to pass to the address bar
        :param tab_index: which tab to open
        """
        logger.warning(
            "In Chrome version 141.0.7390.111, we have noticed clicking on a tab with index might not open the tab on "
            "some devices. Hopefully this will be fixed in future Chrome versions.")
        self._enter_url(url_string, URL_BAR)

    @action(new_tab_state)
    def visit_url_new_tab(self, url_string):
        """
        Creates a new tab and visits the url.
        :param url_string: Url to visit
        """
        self.driver.send_keys(SEARCH_BOX, url_string)
        self.driver.press_enter()

    @action(new_incognito_tab_state)
    def visit_url_incognito(self, url_string: str):
        """
        Opens an incognito tab and enters the url_string to the address bar.
        :param url_string: the input to pass to the address bar
        """
        self._open_settings_pane()
        self.driver.click(NEW_INCOGNITO_TAB_BUTTON)
        self._enter_url(url_string, URL_BAR)

    @action(current_tab_state)
    def bookmark_page(self, tab_index: int):
        """
        Bookmarks the current page.
        :param tab_index: Index of the tab to bookmark.
        :return: True if bookmark has been added, False if it already existed.
        """
        self._open_settings_pane()
        if self.driver.is_present(EDIT_BOOKMARK_BUTTON):
            self.gtl_logger.info("This page was already bookmarked, skipping...")
            return False
        else:
            self.driver.click(BOOKMARK_THIS_PAGE_BUTTON)
            return True

    @action(bookmarks_folder_state)
    def load_first_bookmark(self, folder_name: str):
        """
        Load the first saved bookmark in the specified folder.
        :param folder_name: The name of the folder to load the first bookmark from.
        """
        self.driver.click(FIRST_BOOKMARK)

    @action(current_tab_state)
    def delete_bookmark(self, tab_index: int):
        """
        Delete the current bookmark.
        :param tab_index: Index of the tab to delete the bookmark from.
        :return: True if bookmark has been deleted, False if it wasn't bookmarked.
        """
        self._open_settings_pane()
        if self.driver.is_present(BOOKMARK_THIS_PAGE_BUTTON):
            self.gtl_logger.info("This page was not bookmarked, skipping...")
            return False
        else:
            self.driver.click(EDIT_BOOKMARK_BUTTON)
            self.driver.click(DELETE_BOOKMARK)
            return True

    def _open_settings_pane(self):
        """
        Opens the settings pane. Add a short sleep to ensure the page is actually opened.
        """
        self.driver.click(THREE_DOTS)
        sleep(1)


    def _enter_url(self, url_string: str, url_bar_xpath):
        self.driver.send_keys(url_bar_xpath, url_string)
        self.driver.press_enter()



import unittest

from puma.state_graph.puma_driver import PumaDriver
from puma.state_graph.state import State, ContextualState
from puma.state_graph.state_graph import StateGraphMeta, StateGraph


class TestState(State):
    def __init__(self, id: str, **kwargs):
        super().__init__(**kwargs)
        self.id = id

    def validate(self, driver: PumaDriver) -> bool:
        return True


class TestContextualState(ContextualState):
    def __init__(self, id: str, **kwargs):
        super().__init__(**kwargs)
        self.id = id

    def validate_context(self, driver: PumaDriver) -> bool:
        return True

    def validate(self, driver: PumaDriver) -> bool:
        return True


class TestStateGraphMeta(unittest.TestCase):
    def test_validate_graph_success(self):
        # Create a valid state graph
        state1 = TestState(id="State1", initial_state=True)
        state2 = TestState(id="State2")
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]

        # This should not raise any exceptions
        try:
            StateGraphMeta._validate_graph(states)
        except ValueError:
            self.fail("_validate_graph raised ValueError unexpectedly!")

    def test_validate_graph_no_initial_state(self):
        # Create an invalid state graph with no initial state
        state1 = TestState(id="State1")
        state2 = TestState(id="State2")
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)

    def test_validate_graph_multiple_initial_states(self):
        # Create an invalid state graph with multiple initial states
        state1 = TestState(id="State1", initial_state=True)
        state2 = TestState(id="State2", initial_state=True)
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)

    def test_validate_graph_initial_state_is_contextual(self):
        # Create an invalid state graph where the initial state is a ContextualState
        state1 = TestContextualState(id="State1", initial_state=True)
        state2 = TestState(id="State2")
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)

    def test_validate_graph_unreachable_states(self):
        # Create an invalid state graph with unreachable states
        state1 = TestState(id="State1", initial_state=True)
        state2 = TestState(id="State2")
        state3 = TestState(id="State3")
        state1.to(state2, None)
        state2.to(state1, None)
        state2.to(state3, None)
        states = [state1, state2, state3]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)

    def test_validate_graph_contextual_state_without_parent(self):
        # Create an invalid state graph with a ContextualState without a parent
        state1 = TestState(id="State1", initial_state=True)
        state2 = TestContextualState(id="State2")
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)

    def test_validate_graph_duplicate_state_ids(self):
        # Create an invalid state graph with duplicate state names
        state1 = TestState(id="State1", initial_state=True)
        state2 = TestState(id="State1")
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)

    def test_validate_graph_duplicate_transitions(self):
        # Create an invalid state graph with duplicate transitions
        state1 = TestState(id="State1", initial_state=True)
        state2 = TestState(id="State2")
        state1.to(state2, None)
        state1.to(state2, None)
        state2.to(state1, None)
        states = [state1, state2]
        with self.assertRaises(ValueError):
            StateGraphMeta._validate_graph(states)


class TestStateGraph(unittest.TestCase):
    def test_invalid_package_name(self):
        with self.assertRaises(ValueError) as error:
            graph = StateGraph(device_udid='emulator123', app_package='this is invalid')
        self.assertEquals('The provided package name is invalid: this is invalid', str(error.exception))
        with self.assertRaises(ValueError) as error:
            graph = StateGraph(device_udid='emulator123', app_package='')  # also invalid
        self.assertEquals('The provided package name is invalid: ', str(error.exception))


if __name__ == '__main__':
    unittest.main()



import re
import unittest

import git

from puma import version
from puma.utils import PROJECT_ROOT


def get_current_branch_name() -> str | None:
    """
    Get the name of the current Git branch.
    If the repo is in a detached state it is not in a branch and this method returns None.
    :return: The name of the current branch.
    """
    try:
        repo = git.Repo(search_parent_directories=True)
        return repo.active_branch.name
    except TypeError:
        return None

def extract_issue_number(branch_name) -> str:
    """
    Get the issue number of the current Git branch. Assumes the structure is [ISSUE_NB]-[ISSUE_TITLE], as generated by
    GitHub.
    :param branch_name: Branch name to get the issue number from.
    :return: issue number
    """
    # Use regex to find the issue number at the start of the branch name
    match = re.match(r'^(\d+)-', branch_name)
    if match:
        return match.group(1)  # Return the captured issue number
    else:
        raise ValueError('Invalid issue number, please check that your branch name starts with the issue number and a'
                         ' dash.')

def read_release_notes(file_path: str) -> list[str]:
    """
    Read the content of the release notes file.
    :param file_path: The path to the release notes file.
    :return: The content of the release notes file.
    """
    with open(file_path, 'r') as file:
        return file.readlines()


class TestReleaseNotes(unittest.TestCase):
    def setUp(self):
        self.branch_name = get_current_branch_name()

        # If running on the main branch, skip these tests because branch naming conventions do not apply.
        if self.branch_name == "main":
            self.skipTest("Skipping release-notes tests on the main branch")
        # Also skip if in a detached state
        if self.branch_name is None:
            self.skipTest("Skipping release-notes tests when not on a branch")

        self.issue_number = extract_issue_number(self.branch_name)
        self.release_notes_path = f"{PROJECT_ROOT}/RELEASE_NOTES"
        self.release_notes = read_release_notes(self.release_notes_path)

    def test_branch_in_release_notes(self):
        self.assertIn(f"{self.issue_number}. ",
                      "\n".join(self.release_notes),
                      f"Issue number {self.issue_number} was not found in the release notes, please add an entry "
                      f"for this issue.")

    def test_version_in_release_notes_same_as_setup(self):
        first_line = self.release_notes[0]
        match = re.search(r'(\d+\.\d+\.\d+)', first_line)
        first_version_in_release_notes = match.group(1) if match else None
        self.assertIsNotNone(first_version_in_release_notes,
                             "The first line of the release notes does not contain a semantic version, please "
                             "check this.")
        self.assertEqual(first_version_in_release_notes, version.version,
                         "Version in release notes is not equal to setup version")


if __name__ == '__main__':
    unittest.main()



from puma.state_graph.action import action
from puma.state_graph.state import SimpleState, compose_clicks
from puma.state_graph.state_graph import StateGraph

APPLICATION_PACKAGE = "INSERT YOUR PACKAGE HERE"


# Define custom methods to navigate to a certain state here

class TemplateApp(StateGraph):
    # Define states
    state1 = SimpleState(xpaths=["xpath1", "xpath2"],
                         initial_state=True)
    state2 = SimpleState(xpaths=["xpath1"],
                         parent_state=state1)

    # Define transitions. Only forward transitions are needed, back transitions are added automatically
    state1.to(state2, compose_clicks(['xpath1', 'xpath2']))

    # init
    def __init__(self, device_udid):
        StateGraph.__init__(self, device_udid, APPLICATION_PACKAGE)

    # Define your actions
    @action(state1)
    def action_1(self):
        ...


if __name__ == "__main__":
    app = TemplateApp(device_udid="INSERT YOUR DEVICE ID HERE")
    app.action_1()



from puma.utils.xpath_utils import build_resource_id_xpath_widget, build_resource_id_xpath, \
    build_resource_id_text_xpath_widget, build_resource_id_text_xpath, build_content_desc_xpath_widget, \
    build_text_xpath, build_content_desc_xpath, build_text_xpath_widget


def build_wa_resource_id_xpath_widget(widget_type: str, resource_id: str) -> str:
    return build_resource_id_xpath_widget(widget_type, WHATSAPP_PACKAGE, resource_id)


def build_wa_resource_id_xpath(resource_id: str) -> str:
    return build_resource_id_xpath(WHATSAPP_PACKAGE, resource_id)


def build_wa_resource_id_text_xpath_widget(widget_type: str, resource_id: str, text: str) -> str:
    return build_resource_id_text_xpath_widget(widget_type, WHATSAPP_PACKAGE, resource_id, text)


def build_wa_resource_id_text_xpath(resource_id: str, text: str) -> str:
    return build_resource_id_text_xpath(WHATSAPP_PACKAGE, resource_id, text)


WHATSAPP_PACKAGE = 'com.whatsapp'

CONVERSATIONS_WHATSAPP_LOGO = build_wa_resource_id_xpath_widget('ImageView', 'toolbar_logo')
CONVERSATIONS_NEW_CHAT_OR_SEND_MESSAGE = f'{build_content_desc_xpath_widget("ImageButton", "New chat")} | {build_content_desc_xpath_widget("Button", "Send message")}'
CONVERSATIONS_HOME_ROOT_FRAME = build_wa_resource_id_xpath_widget('FrameLayout', 'root_view')
CONVERSATIONS_MENUITEM_ARCHIVE = build_wa_resource_id_xpath('menuitem_conversations_archive')
CONVERSATIONS_ARCHIVED = f'//*[contains(@text,"archived") or @resource-id="{WHATSAPP_PACKAGE}:id/fab"]'
CONVERSATIONS_GROUP_NAME = build_wa_resource_id_xpath('group_name')
CONVERSATIONS_NEW_GROUP = build_text_xpath('New group')
CONVERSATIONS_CHAT_ABLE_CONTACT = build_wa_resource_id_text_xpath('chat_able_contacts_row_name', '{receiver}')
CONVERSATIONS_NEW_BROADCAST_TITLE = build_wa_resource_id_text_xpath('title', 'New broadcast')
CONVERSATIONS_ROW_BY_SUBJECT = f'//*[contains(@resource-id,"{WHATSAPP_PACKAGE}:id/conversations_row_contact_name") and @text="{{conversation}}"]'

SETTINGS_QR = build_wa_resource_id_xpath_widget('ImageView', 'profile_info_qr_code')
SETTINGS_ACCOUNT_SWITCH = build_wa_resource_id_xpath_widget('ImageView', 'account_switcher_button')

PROFILE_PROFILE_PICTURE = build_wa_resource_id_xpath_widget('ImageView', 'photo_btn')
PROFILE_NAME = build_wa_resource_id_text_xpath_widget('Button', 'profile_settings_row_text', 'Name')
PROFILE_PHONE = build_wa_resource_id_text_xpath_widget('Button', 'profile_settings_row_text', 'Phone')
PROFILE_INFO_EDIT_BUTTON = build_wa_resource_id_xpath_widget('Button', 'profile_info_edit_btn')
PROFILE_GALLERY = build_text_xpath('Gallery')
PROFILE_FOLDERS = build_content_desc_xpath_widget('ImageButton', 'Folders')
PROFILE_SAVE_BUTTON = build_wa_resource_id_xpath('save_button')
PROFILE_STATUS_EDIT_ICON = build_wa_resource_id_xpath('status_tv_edit_icon')
PROFILE_INFO_STATUS_CARD = build_wa_resource_id_xpath('profile_info_status_card')

NEW_CHAT_HEADER = build_text_xpath_widget('TextView', 'New chat')
NEW_CHAT_NEW_GROUP = build_wa_resource_id_text_xpath_widget('TextView', 'contactpicker_row_name', 'New group')
NEW_CHAT_NEW_CONTACT = build_wa_resource_id_text_xpath_widget('TextView', 'contactpicker_row_name', 'New contact')
NEW_CHAT_NEW_COMMUNITY = build_wa_resource_id_text_xpath_widget('TextView', 'contactpicker_row_name', 'New community')

CHAT_ROOT_LAYOUT = build_wa_resource_id_xpath_widget('LinearLayout', 'conversation_root_layout')
CHAT_CONTACT_HEADER = build_wa_resource_id_xpath_widget('TextView', 'conversation_contact_name')
CHAT_CONTACT_HEADER_TEXT = f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/conversation_contact_name" and contains(lower-case(@text), "{{conversation}}")]'
CHAT_CONTACT_HEADER_WITH_NAME = build_wa_resource_id_text_xpath_widget('TextView', 'conversation_contact_name', '{conversation}')
CHAT_DELETE_BUTTON = build_content_desc_xpath('Delete')
CHAT_DELETE_FOR_EVERYONE = f'//*[@resource-id="{WHATSAPP_PACKAGE}:id/buttonPanel"]//*[@text="Delete for everyone"]'
CHAT_REPLY = build_content_desc_xpath('Reply')
CHAT_VIEW_ONCE_MEDIA = '//*[contains(@resource-id, "view_once_media")]'
CHAT_VOICE_NOTE_BUTTON = build_wa_resource_id_xpath('voice_note_btn')
CHAT_STICKER = build_wa_resource_id_xpath('sticker')
CHAT_STICKERS = build_wa_resource_id_xpath('stickers')
CHAT_EMOJI = build_wa_resource_id_xpath('emoji')
CHAT_EMOJIS = build_wa_resource_id_xpath('emojis')
CHAT_EMOJI_PICKER = build_wa_resource_id_xpath('emoji_picker_btn')
CHAT_DIRECTORY_NAME = '//android.widget.TextView[@text="{directory_name}"]'
CHAT_DIRECTORY_MEDIA_BY_INDEX = '//androidx.compose.ui.platform.ComposeView/android.view.View/android.view.View/android.view.View[4]/android.view.View[{index}]/android.view.View[2]/android.view.View'
CHAT_POPUP_BUTTON_OK = build_wa_resource_id_xpath_widget('Button', 'vo_sp_bottom_sheet_ok_button')
CHAT_CAPTION_TEXT_BOX = build_wa_resource_id_xpath('caption')
CHAT_SEND_MEDIA_VIEW_ONCE = build_wa_resource_id_xpath('view_once_toggle')
CHAT_CONTACT_NAME = build_wa_resource_id_text_xpath_widget('TextView', 'name', '{contact_name}')
CHAT_ATTACH_SEND_BUTTON = build_wa_resource_id_xpath('send_btn')
CHAT_ATTACH_CONTACT_BUTTON = build_wa_resource_id_xpath('pickfiletype_contact_holder')
CHAT_ATTACH_GALLERY_BUTTON = build_wa_resource_id_xpath('pickfiletype_gallery_holder')
CHAT_GALLERY_FOLDERS_BUTTON = build_content_desc_xpath_widget('ImageButton', 'Folders')
CHAT_FIRST_MEDIA_IN_FOLDER = '//androidx.compose.ui.platform.ComposeView/android.view.View/android.view.View/android.view.View[5]/android.view.View[3]/android.widget.Button'
CHAT_STOP_SHARING = build_text_xpath('Stop sharing')
CHAT_MESSAGE_BY_CONTENT = f"//*[@resource-id='{WHATSAPP_PACKAGE}:id/conversation_text_row']//*[contains(@text,'{{message_contains}}')]"
CHAT_MESSAGE_BY_CONTENT_AND_STATE = (f'(//android.widget.FrameLayout['
                              f'@resource-id="{WHATSAPP_PACKAGE}:id/conversation_text_row"'
                              f' and .//android.widget.TextView[@resource-id="com.whatsapp:id/message_text" and @text="{{message_text}}"]'
                              f' and .//android.widget.ImageView[@content-desc="{{state}}"]])')
CHAT_FORWARD_MESSAGE = f"//*[@resource-id='{WHATSAPP_PACKAGE}:id/action_mode_bar']//*[@content-desc='Forward']"
CHAT_FORWARD_CONTACT_BY_NAME = f"//*[@resource-id='{WHATSAPP_PACKAGE}:id/contact_list']//*[@text='{{to_chat}}']"
CHAT_MENTION_SUGGESTIONS = build_wa_resource_id_xpath_widget('ImageView', 'contact_photo')
CHAT_ATTACH_BUTTON = build_wa_resource_id_xpath('input_attach_button')
CHAT_ATTACH_LOCATION_BUTTON = build_wa_resource_id_xpath_widget('Button', 'pickfiletype_location_holder')

SEND_LOCATION_HEADER = f'//android.view.ViewGroup[@resource-id="{WHATSAPP_PACKAGE}:id/toolbar"]/android.widget.TextView[@text="Send location"]'
SEND_LOCATION_CURRENT_LOCATION = build_wa_resource_id_xpath_widget('FrameLayout', 'send_current_location_btn')
SEND_LOCATION_CURRENT_LOCATION_BUTTON = build_wa_resource_id_xpath('send_current_location_btn')
SEND_LOCATION_LIVE_LOCATION = build_wa_resource_id_xpath_widget('FrameLayout', 'live_location_btn')
SEND_LOCATION_LIVE_LOCATION_BUTTON = build_wa_resource_id_xpath('live_location_btn')
SEND_LOCATION_LIVE_LOCATION_DIALOG = build_wa_resource_id_xpath_widget('LinearLayout', 'location_new_user_dialog_container')
SEND_LOCATION_POPUP_CONTINUE = build_text_xpath_widget('Button', 'Continue')
SEND_LOCATION_CAPTION = build_wa_resource_id_xpath('comment')

CHAT_SETTINGS_CONTACT_NAME = (f'{build_wa_resource_id_xpath_widget("TextView", "contact_title")} | '
                              f'{build_wa_resource_id_xpath_widget("TextView", "business_title")} | '
                              f'{build_wa_resource_id_xpath_widget("TextView", "group_title")}')
CHAT_SETTINGS_CONTACT_NAME_TEXT = (f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/contact_title" and contains(lower-case(@text), "{{conversation}}")] | '
                                   f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/business_title" and contains(lower-case(@text), "{{conversation}}")] | '
                                   f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/group_title" and contains(lower-case(@text), "{{conversation}}")]')
CHAT_SETTINGS_NOTIFICATIONS = build_wa_resource_id_xpath_widget('LinearLayout', 'notifications_and_sounds_layout')
CHAT_SETTINGS_MEDIA_VISIBILITY = build_wa_resource_id_xpath_widget('Button', 'media_visibility_layout')
CHAT_SETTINGS_ANY_MEMBER = f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/name"]'
CHAT_SETTINGS_MEMBER = build_wa_resource_id_text_xpath('name', '{member}')
CHAT_SETTINGS_REMOVE_MEMBER = "//*[starts-with(@text, 'Remove')]"
CHAT_SETTINGS_EXIT_GROUP_BUTTON = build_text_xpath_widget('Button', 'Exit group')
CHAT_SETTINGS_EXIT_GROUP_LIST_ITEM = build_wa_resource_id_text_xpath('list_item_title', 'Exit group')
CHAT_SETTINGS_DISAPPEARING_MESSAGES = f'//*[@resource-id="{WHATSAPP_PACKAGE}:id/list_item_title" and @text="Disappearing messages"]'
CHAT_SETTINGS_OK_BUTTON = build_text_xpath_widget('Button', 'OK')
CHAT_SETTINGS_CONTAINS_DELETE_GROUP = '//*[contains(@text,"Delete group")]'

RADIO_BUTTON_24_HOURS = build_text_xpath_widget('RadioButton', '24 hours')
RADIO_BUTTON_OFF = build_text_xpath_widget('RadioButton', 'Off')

UPDATES_HEADER = f'//android.view.ViewGroup[@resource-id="{WHATSAPP_PACKAGE}:id/toolbar"]/android.widget.TextView[@text="Updates"]'
UPDATES_STATUS_HEADER = build_wa_resource_id_text_xpath_widget('TextView', 'header_textview', 'Status')
UPDATES_NEW_STATUS = build_content_desc_xpath_widget('ImageButton', 'New status update')
UPDATES_EDIT_CAPTION = build_wa_resource_id_xpath_widget('EditText', 'caption')
UPDATES_SHUTTER = build_wa_resource_id_xpath_widget('ImageView', 'shutter')
UPDATES_CAMERA_BUTTON = build_content_desc_xpath_widget('Button', 'Camera')

CALLS_START_CALL = build_content_desc_xpath_widget('ImageButton', 'New call')
CALLS_HEADER = f'//android.view.ViewGroup[@resource-id="{WHATSAPP_PACKAGE}:id/toolbar"]/android.widget.TextView[@text="Calls"]'

CALL_CONTACT_HEADER = build_wa_resource_id_xpath_widget('TextView', 'title')
CALL_CONTACT_HEADER_TEXT = f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/title" and contains(lower-case(@text), "{{conversation}}")]'
CALL_TAB_SEARCH_BUTTON = build_content_desc_xpath_widget('ImageButton', 'Search')
CALL_SCREEN_BACKGROUND = build_wa_resource_id_xpath_widget('RelativeLayout', 'call_screen')
CALL_CONTACT_ROW = f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/contact_name" and @text="{{conversation}}"]'
CALL_END_CALL_BUTTON = ('//*[@content-desc="Leave call" or '
                        f'@resource-id="{WHATSAPP_PACKAGE}:id/end_call_button" or '
                        f'@resource-id="{WHATSAPP_PACKAGE}:id/footer_end_call_btn"]')

VOICE_CALL_START_BUTTON = build_wa_resource_id_xpath('voice_call')
VIDEO_CALL_START_BUTTON = build_wa_resource_id_xpath('video_call')
VOICE_CALL_CAMERA_BUTTON = f'//android.widget.Button[@content-desc="Turn camera on" and @resource-id="{WHATSAPP_PACKAGE}:id/camera_button"]'
VIDEO_CALL_CAMERA_BUTTON = f'//android.widget.Button[@content-desc="Turn camera off" and @resource-id="{WHATSAPP_PACKAGE}:id/camera_button"]'
VIDEO_CALL_SWITCH_CAMERA = build_wa_resource_id_xpath_widget('Button', 'calling_camera_switch_wds_button')

RECEIVE_CALL_ANSWER_BUTTON = "//android.widget.Button[@content-desc='Answer' or @content-desc='Video']"
RECEIVE_CALL_DECLINE_BUTTON = build_content_desc_xpath_widget('Button', 'Decline')

TEXT_ENTRY = build_wa_resource_id_xpath('entry')
TEXT_VIEWS = '//android.widget.TextView'

OPEN_SETTINGS_BY_TITLE = build_text_xpath_widget('TextView', 'Settings')
PROFILE_INFO = build_wa_resource_id_xpath_widget('TextView', 'profile_info_name')

HAMBURGER_MENU = build_content_desc_xpath_widget('ImageView', 'More options')
SEARCH_BAR = build_wa_resource_id_xpath_widget('EditText', 'search_view_edit_text')
CONVERSATIONS_CHAT_TAB = build_content_desc_xpath_widget('FrameLayout', 'Chats')
UPDATES_TAB = build_wa_resource_id_text_xpath_widget('TextView', 'navigation_bar_item_small_label_view', 'Updates')
CALLS_TAB = build_wa_resource_id_text_xpath_widget('TextView', 'navigation_bar_item_small_label_view', 'Calls')

SEND = build_wa_resource_id_xpath('send')
SEND_CONTENT = build_content_desc_xpath_widget('ImageButton', 'Send')
SEND_RESOURCE = build_wa_resource_id_xpath_widget('ImageButton', 'send')
STOP_BUTTON = build_content_desc_xpath_widget('Button', 'Stop')
NEXT_BUTTON = build_wa_resource_id_xpath('next_btn')
EDIT_TEXT = build_wa_resource_id_xpath('edit_text')
OK_BUTTON = build_wa_resource_id_xpath('ok_btn')



from puma.apps.android.whatsapp.xpaths import *
from puma.state_graph.puma_driver import PumaDriver
from puma.state_graph.state import SimpleState, ContextualState, State


class WhatsAppChatState(SimpleState, ContextualState):
    """
    A state representing a chat screen in the application.

    This class extends both SimpleState and ContextualState to represent a chat screen
    and validate its context based on the conversation name.
    """

    def __init__(self, parent_state: State):
        """
        Initializes the ChatState with a parent state.

        :param parent_state: The parent state of this chat state.
        """
        super().__init__(xpaths=[CHAT_CONTACT_HEADER,
                                 CHAT_ROOT_LAYOUT],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        """
        Validates the context of the chat state.

        This method checks if the current chat screen matches the expected conversation name.

        :param driver: The PumaDriver instance used to interact with the application.
        :param conversation: The name of the conversation to validate against.
        :return: True if the context is valid, False otherwise.
        """
        if not conversation:
            return True

        return driver.is_present(CHAT_CONTACT_HEADER_TEXT.format(conversation=conversation.lower()))


    @staticmethod
    def open_chat_settings(driver: PumaDriver, conversation: str):
        driver.click(CHAT_CONTACT_HEADER_WITH_NAME.format(conversation=conversation))


class WhatsAppChatSettingsState(SimpleState, ContextualState):
    """
    A state representing a chat settings screen in the application.

    This class extends both SimpleState and ContextualState to represent a chat settings screen
    and validate its context based on the contact or group name.
    """

    def __init__(self, parent_state: State):
        """
        Initializes the ChatSettingsState with a parent state.

        :param parent_state: The parent state of this chat state.
        """
        super().__init__(xpaths=[CHAT_SETTINGS_CONTACT_NAME,
                                 CHAT_SETTINGS_NOTIFICATIONS,
                                 CHAT_SETTINGS_MEDIA_VISIBILITY],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        """
        Validates the context of the chat settings state.

        This method checks if the current chat settings screen matches the expected contact or group name.

        :param driver: The PumaDriver instance used to interact with the application.
        :param conversation: The name of the conversation to validate against.
        :return: True if the context is valid, False otherwise.
        """
        if not conversation:
            return True

        return driver.is_present(CHAT_SETTINGS_CONTACT_NAME_TEXT.format(conversation=conversation.lower()))


class WhatsAppVoiceCallState(SimpleState, ContextualState):
    """
    A state representing a call screen in the application.

    This class extends both SimpleState and ContextualState to represent a call screen
    and validate its context based on the contact.
    """

    def __init__(self, parent_state: State):
        """
        Initializes the CallState with a parent state.

        :param parent_state: The parent state of this call state.
        """
        super().__init__(xpaths=[CALL_END_CALL_BUTTON,
                                 CALL_SCREEN_BACKGROUND,
                                 VOICE_CALL_CAMERA_BUTTON],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        """
        Validates the context of the call state.

        This method checks if the current call screen matches the expected contact name.

        :param driver: The PumaDriver instance used to interact with the application.
        :param conversation: The name of the call recipient to validate against.
        :return: True if the context is valid, False otherwise.
        """
        if not conversation:
            return True

        return driver.is_present(CALL_CONTACT_HEADER_TEXT.format(conversation=conversation.lower()))


class WhatsAppVideoCallState(SimpleState, ContextualState):
    """
    A state representing a call screen in the application.

    This class extends both SimpleState and ContextualState to represent a call screen
    and validate its context based on the contact.
    """

    def __init__(self, parent_state: State):
        """
        Initializes the CallState with a parent state.

        :param parent_state: The parent state of this call state.
        """
        super().__init__(xpaths=[CALL_END_CALL_BUTTON,
                                 CALL_SCREEN_BACKGROUND,
                                 VIDEO_CALL_CAMERA_BUTTON,
                                 VIDEO_CALL_SWITCH_CAMERA],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        """
        Validates the context of the call state.

        This method checks if the current call screen matches the expected contact name.

        :param driver: The PumaDriver instance used to interact with the application.
        :param conversation: The name of the call recipient to validate against.
        :return: True if the context is valid, False otherwise.
        """
        if not conversation:
            return True

        return driver.is_present(CALL_CONTACT_HEADER_TEXT.format(conversation=conversation.lower()))



from puma.apps.android.google_chrome import logger
from puma.apps.android.google_chrome.xpaths import TAB_SWITCH_BUTTON, TAB_LIST, URL_BAR, \
    NEW_TAB_FROM_CURRENT_TAB, BOOKMARKS_SORT_VIEW, BOOKMARKS_CREATE_FOLDER, BOOKMARKS_GO_BACK
from puma.state_graph.puma_driver import PumaDriver
from puma.state_graph.state import SimpleState, ContextualState, compose_clicks, TransitionError


class BookmarksFolder(SimpleState, ContextualState):
    def __init__(self, parent_state):
        """
        Initializes the BookmarksFolder state with a given parent state.

        :param parent_state: The parent state of this state
        """
        super().__init__(xpaths=[BOOKMARKS_SORT_VIEW, BOOKMARKS_GO_BACK, BOOKMARKS_CREATE_FOLDER],
                         parent_state=parent_state,
                         parent_state_transition=compose_clicks([BOOKMARKS_GO_BACK]))

    def validate_context(self, driver: PumaDriver, folder_name: str = None) -> bool:
        """
        Validates if we are in the correct bookmarks folder.
        :param driver: Puma driver
        :param folder_name: Name of the bookmarks folder
        :return: boolean
        """
        if folder_name is None:
            return True
        return driver.is_present(f'//android.view.ViewGroup[@resource-id="com.android.chrome:id/action_bar"]//android.widget.TextView[@text="{folder_name}"]')

    @staticmethod
    def go_to_bookmarks_folder(driver: PumaDriver, folder_name: str):
        driver.click(f'//android.widget.TextView[@resource-id="com.android.chrome:id/title" and @text="{folder_name}"]')


class CurrentTab(SimpleState, ContextualState):
    """
    A state representing an existing tab in the application.

    This class extends both SimpleState and ContextualState to represent a tab screen. The contextual state is an outlier,
    see the validate_context method.
    """

    def __init__(self, parent_state):
        """
        Initializes the Current Tab state with a given parent state.

        :param parent_state: The parent state of this current tab state.
        """
        super().__init__(
            xpaths=[URL_BAR, NEW_TAB_FROM_CURRENT_TAB],
            parent_state=parent_state,
            parent_state_transition=compose_clicks([TAB_SWITCH_BUTTON]))
        # Keep a dict that tracks which tab indices were opened last on which device. See validate_context()
        self.last_opened = {}


    def validate_context(self, driver: PumaDriver, tab_index: int = None) -> bool:
        """
        We can not validate if we are in the nth tab from the tab overview, while in the tab contextual state, as this
        index is not available in the state.
        Therefore, we opted to store the last opened tab index for each device in this state. This means that the
        contextual state is not actually verified against the UI, but against the tab index saved in the last opened
        dict. When switching between two tabs, this works as long as the user does not interrupt Puma during these UI
        actions.
        #TODO document this situation in the CONTRIBUTING
        :param driver: Puma driver
        :param tab_index: Index of the tab last opened
        :return: boolean
        """
        if not tab_index:
            return True
        try:
            return self.last_opened[driver.udid] == tab_index
        except KeyError:
            return False

    def switch_to_tab(self, driver: PumaDriver, tab_index: int):
        """
        Navigate to the tab at the specified index in the tab overview. This method also stores which tab index was
        opened on which device, enabling contextual validation in validate_context().

        :param driver: The PumaDriver instance used to interact with the application
        :param tab_index: Index of the tab to navigate to
        """
        tab_content_view = f'({TAB_LIST}//*[@resource-id="com.android.chrome:id/content_view"])[{tab_index}]'
        if driver.is_present(f'{tab_content_view}'
                             f'//*'
                             f'[@resource-id="com.android.chrome:id/tab_title" and @text="New tab"]'):
            logger.error(f"The tab at index {tab_index} is a new tab, not an existing tab. Use the new tab action "
                         f"instead.")
            raise TransitionError()
        driver.click(tab_content_view)
        self.last_opened[driver.udid] = tab_index



import time

from puma.apps.android.telegram import logger
from puma.apps.android.telegram.xpaths import *
from puma.state_graph.action import action
from puma.state_graph.puma_driver import supported_version, PumaDriver, PumaClickException
from puma.state_graph.state import SimpleState, ContextualState, State, compose_clicks
from puma.state_graph.state_graph import StateGraph

TELEGRAM_PACKAGE = 'org.telegram.messenger'
TELEGRAM_WEB_PACKAGE = 'org.telegram.messenger.web'


class TeleGramChatState(SimpleState, ContextualState):
    """
    State defining a Telegram chat screen (group or 1 on 1)
    """

    def __init__(self, parent_state: State):
        super().__init__([CHAT_STATE_MESSAGE_TEXTBOX, CHAT_STATE_GIF_BUTTON],
                         parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        if not conversation:
            return True
        return driver.is_present(CHAT_STATE_CONVERSATION_NAME.format(conversation=conversation))

    @staticmethod
    def go_to_chat(driver: PumaDriver, conversation: str):
        if not conversation:
            raise ValueError(f'Cannot open a conversation without a conversation name')
        driver.click(CHAT_OVERVIEW_SEARCH_BUTTON)
        driver.send_keys(SEARCH_INPUT_FIELD, conversation)
        driver.click(FIRST_SEARCH_HIT.format(conversation=conversation))


class TelegramChatSettingsState(SimpleState, ContextualState):
    """
    State defining the chat settings screen of a Telegram conversation (group or 1 on 1).
    """

    def __init__(self, parent_state: State):
        super().__init__(
            [CHAT_SETTINGS_STATE_BACK, CHAT_SETTINGS_STATE_THREE_DOTS, CHAT_SETTINGS_STATE_CONVERSATION_NAME],
            parent_state=parent_state,
            # invalid_xpaths is needed because this state was confused with the send_from_gallery state
            invalid_xpaths=[SEND_FROM_GALLERY_MEDIA_SWITCH.format(index=1)])

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        if conversation is None:
            return True
        return driver.is_present(CHAT_SETTINGS_STATE_CONVERSATION_NAME_CONTEXT.format(conversation=conversation))

    @staticmethod
    def can_add_members(driver: PumaDriver):
        return driver.is_present(CHAT_SETTINGS_STATE_ADD_MEMBERS)

    @staticmethod
    def open_chat_settings(driver: PumaDriver, conversation: str):
        driver.click(CHAT_STATE_CONVERSATION_NAME.format(conversation=conversation))


@supported_version("12.0.1")
class Telegram(StateGraph):
    """
    A class representing a state graph for managing UI states and transitions in the Telegram Messenger application.

    This class uses a state machine approach to manage transitions between different states
    of the Telegram user interface. It provides methods to navigate between states, validate states,
    and handle unexpected states or errors.
    """
    conversations_state = SimpleState(
        [CHAT_OVERVIEW_NEW_MESSAGE_BUTTON, CHAT_OVERVIEW_SEARCH_BUTTON, CHAT_OVERVIEW_NAV_MENU_BUTTON],
        initial_state=True)
    chat_state = TeleGramChatState(parent_state=conversations_state)
    chat_settings_state = TelegramChatSettingsState(parent_state=chat_state)
    call_state = SimpleState([CALL_STATE_END_CALL_BUTTON, CALL_STATE_MUTE_BUTTON, CALL_STATE_SPEAKER_BUTTON],
                             parent_state=chat_state)
    send_media_state = SimpleState(
        [SEND_MEDIA_STATE_INSTANT_CAMERA_BUTTON, SEND_MEDIA_STATE_GALLERY_BUTTON, SEND_MEDIA_STATE_FILE_BUTTON,
         SEND_MEDIA_STATE_LOCATION_BUTTON],
        parent_state=chat_state)
    send_from_gallery_state = SimpleState(
        [SEND_FROM_GALLERY_STATE_BACK_BUTTON, SEND_FROM_GALLERY_FOLDER_PICKER,
         SEND_FROM_GALLERY_THREE_DOTS_BUTTON, SEND_FROM_GALLERY_MEDIA_SWITCH.format(index=1)],
        parent_state=chat_state)  # pressing back goes to the chat state
    new_message_state = SimpleState(
        [NEW_MESSAGE_STATE_NEW_GROUP_BUTTON, NEW_MESSAGE_STATE_NEW_CONTACT_BUTTON, NEW_MESSAGE_STATE_NEW_CHANNEL_BUTTON,
         NEW_MESSAGE_STATE_CREATE_NEW_CONTACT_BUTTON],
        parent_state=conversations_state)

    conversations_state.to(chat_state, TeleGramChatState.go_to_chat)
    chat_state.to(call_state, compose_clicks([CHAT_STATE_CALL_BUTTON], name="press_call_button"))
    chat_state.to(chat_settings_state, TelegramChatSettingsState.open_chat_settings)
    chat_state.to(send_media_state, compose_clicks([CHAT_STATE_MEDIA_BUTTON], name="press_attachment_button"))
    send_media_state.to(send_from_gallery_state,
                        compose_clicks([SEND_MEDIA_STATE_GALLERY_BUTTON], name='press_gallery_button'))
    conversations_state.to(new_message_state,
                           compose_clicks([CHAT_OVERVIEW_NEW_MESSAGE_BUTTON], name='press_new_message_button'))

    def __init__(self, device_udid, telegram_web_version: bool = False):
        """
        Initializes Telegram messenger with a device UDID.

        :param device_udid: The unique device identifier for the Android device.
        """
        package = TELEGRAM_WEB_PACKAGE if telegram_web_version else TELEGRAM_PACKAGE
        StateGraph.__init__(self, device_udid, package)

    @action(chat_state)
    def send_message(self, message: str, conversation: str = None):
        """
        Sends a regular text based message.
        Can be done in a given conversation or the current conversation.

        :param message: the message to send
        :param conversation: the conversation (group or 1 on 1) to send the message in. Not needed when already in a conversation.
        """
        # self.driver.click(CHAT_STATE_MESSAGE_TEXTBOX)
        self.driver.send_keys(CHAT_STATE_MESSAGE_TEXTBOX, message)
        # send_button = self.driver.get_element('//android.view.View[@content-desc="Send"]')
        # Telegram has a horrible UI; the bounds of the send button are defined incorrectly so we need to click off-center
        self.driver.click(CHAT_STATE_SEND_BUTTON, 0.8, 0.5)

    @action(chat_state)
    def send_voice_message(self, duration: int, conversation: str = None):
        """
        Sends a voice message. This is a short audio clip that can be recorded in the chat screen.
        Can be done in a given conversation or the current conversation.

        :param duration: the duration of the voice message.
        :param conversation: the conversation (group or 1 on 1) to send the message in. Not needed when already in a conversation.
        """
        if not self._in_voice_message_mode():
            self.driver.click(CHAT_STATE_RECORD_VIDEO_OR_AUDIO_MESSAGE)
        self.driver.long_click_element(CHAT_STATE_RECORD_VIDEO_OR_AUDIO_MESSAGE, duration)

    @action(chat_state)
    def send_video_message(self, duration: int, conversation: str = None):
        """
        Sends a video message. This is a short video clip that can be recorded in the chat screen.
        Can be done in a given conversation or the current conversation.

        :param duration: the duration of the video message.
        :param conversation: the conversation (group or 1 on 1) to send the message in. Not needed when already in a conversation.
        """
        if self._in_voice_message_mode():
            self.driver.click(CHAT_STATE_RECORD_VIDEO_OR_AUDIO_MESSAGE)
        self.driver.long_click_element(CHAT_STATE_RECORD_VIDEO_OR_AUDIO_MESSAGE, duration)

    @action(send_from_gallery_state, end_state=chat_state)
    def send_media_from_gallery(self, media_index: int | list[int],
                                caption: str = None,
                                folder: str | int = None,
                                conversation: str = None):
        """
        Sends one or more video or picture. This is a short audio clip that can be recorded in the chat screen.
        Can be done in a given conversation or the current conversation.

        :param media_index: the index or indices of the media files to send. This is the 1-based index of the thumbnails
        seen in the media picker.
        :param caption: An optional text message attached to the media file(s)
        :param folder: The name of the folder in which the files are stored. Optional: if none given the default
        'Gallery' section is used.
        :param conversation: the conversation (group or 1 on 1) to send the media in. Not needed when already in a
        conversation.
        """
        if not media_index:
            raise ValueError(f'Cannot send media from gallery without a proper set of indices')
        # open folder if needed
        if folder:
            self.driver.click(SEND_FROM_GALLERY_FOLDER_PICKER)
            if isinstance(folder, str):
                logger.warning(
                    f'Using OCR to click on media folder {folder}. OCR is unreliable, if possible use the folder index number!')
                time.sleep(1)
                self.driver.click_text_ocr(folder)
            else:
                self.driver.click(f'//android.widget.LinearLayout/android.view.View[{folder}]')
        # select picture/video
        if isinstance(media_index, int):
            media_index = [media_index]
        clicked = 0
        for i in media_index:
            try:
                if i in [2, 3]:  # these switches can be obscured by other UI elements. We do a long press for those
                    self.driver.long_click_element(f'{SEND_FROM_GALLERY_MEDIA_SWITCH.format(index=i)}/..')
                else:
                    self.driver.click(SEND_FROM_GALLERY_MEDIA_SWITCH.format(index=i))
                self.gtl_logger.info(f'Selected gallery item with index {i}')
                clicked += 1
            except PumaClickException as e:
                logger.warning(f'Could not select media with index {i}. Are enough media files present?')
        if clicked == 0:
            raise PumaClickException(
                f'Tried to select gallery items with indexes {media_index}, but could not select any of them.')
        if caption:
            self.driver.send_keys('//android.widget.EditText[@text]', caption)
            self.gtl_logger.info(f'Entered caption {caption}')
        # send button
        self.driver.click('//android.widget.Button')
        self.gtl_logger.info(f'Pressed send button')

    @action(chat_state, end_state=call_state)
    def start_call(self, conversation: str = None):
        """
        Starts a voice call in the current or given conversation.

        :param conversation: the conversation (group or 1 on 1) to make the call in. Not needed when already in a
        conversation.
        """
        self.driver.click(CHAT_STATE_CALL_BUTTON)

    def answer_call(self):
        """
        Answer when receiving a call via Telegram.
        """
        self.driver.open_notifications()
        ANSWER_CALL_NOTIFICATION_BUTTON = '//android.widget.Button[lower-case(@content-desc)="answer"]'
        if not self.driver.is_present(ANSWER_CALL_NOTIFICATION_BUTTON, implicit_wait=5):
            self.driver.back()
            raise PumaClickException('Tried to answer call, but couldn\'t find notification to answer any calls')
        self.driver.click(ANSWER_CALL_NOTIFICATION_BUTTON)

    @action(call_state)
    def mute_mic(self):
        """
        Mutes or unmutes the microphone when in a call.
        """
        self.driver.click(CALL_STATE_MUTE_BUTTON)

    def end_call(self):
        """
        Ends the current Telegram call. If no call is in progress, this will log a warning but not raise an error.
        """
        if self.current_state != Telegram.call_state:
            self.gtl_logger.warning(f'Tried to end a call, but no call was in progress')
        self.driver.click(CALL_STATE_END_CALL_BUTTON)

    @action(send_media_state, end_state=chat_state)
    def send_current_location(self, conversation: str = None):
        """
        Sends the current location.
        Can be done in a given conversation or the current conversation.

        :param conversation: the conversation (group or 1 on 1) to send the message in. Not needed when already in a conversation.
        """
        self.driver.click(SEND_MEDIA_STATE_LOCATION_BUTTON)
        self.driver.click(SEND_MEDIA_STATE_MY_LOCATION_BUTTON)
        self.driver.click(SEND_MEDIA_STATE_SEND_CURRENT_LOCATION_BUTTON)

    @action(send_media_state, end_state=chat_state)
    def send_live_location(self, conversation: str = None, duration_option: int | str = 1):
        """
        Sends a live location, updating the chat participants of the location for a given amount of time.
        Can be done in a given conversation or the current conversation.

        :param duration_option: the duration option to pick. Typically, the UI offers the choices of 15m, 1 hour and 8
        hours. You can either the index of the option (the first option has index 1) or a string that occurs in the UI
        text (e.g. "15 min"). The latter might change, so an index is probably more stable.
        :param conversation: the conversation (group or 1 on 1) to send the message in. Not needed when already in a conversation.
        """
        self.driver.click(SEND_MEDIA_STATE_LOCATION_BUTTON)
        if not self.driver.is_present(SEND_MEDIA_STATE_LIVE_LOCATION_BUTTON, implicit_wait=1):
            self.gtl_logger.warning(f'Could not share live location, was live location already being shared?')
            self.driver.back()
            return
        self.driver.click(SEND_MEDIA_STATE_LIVE_LOCATION_BUTTON)
        if isinstance(duration_option, int):
            self.driver.click(SEND_MEDIA_STATE_LIVE_LOCATION_DURATION_OPTION_INDEX.format(duration=duration_option))
        else:
            xpath = SEND_MEDIA_STATE_LIVE_LOCATION_DURATION_OPTION_TEXT.format(duration=duration_option)
            if not self.driver.is_present(xpath, implicit_wait=1):
                raise PumaClickException(
                    f'Could not select option for live location sharing with name "{duration_option}". Is this option present in the UI?')
            self.driver.click(xpath)
        self.driver.click(SEND_MEDIA_STATE_LIVE_LOCATION_SHARE_BUTTON)

    @action(chat_state)
    def stop_live_location_sharing(self, conversation: str = None):
        """
        Stops live location sharing in a conversation.

        :param conversation: the conversation (group or 1 on 1) to send the message in. Not needed when already in a conversation.
        """
        if not self.driver.is_present(CHAT_STATE_STOP_LIVE_LOCATION_SHARING_BUTTON):
            self.gtl_logger.warning(f'Could not stop sharing live location as it wasn\'t shared.')
            return
        self.driver.click(CHAT_STATE_STOP_LIVE_LOCATION_SHARING_BUTTON)
        self.driver.click(CHAT_STATE_STOP_LIVE_LOCATION_CONFIRM_BUTTON)

    @action(new_message_state, end_state=chat_state)
    def create_group(self, group_name: str, members: list[str] = [], auto_delete: int = None):
        """
        Create a new chat group.

        :param group_name: the name of the new chat group
        :param members: Optional: a list of members to add when creating the group.
        :param auto_delete: optional: select the auto-delete option to use (1-based index). Off by default.
        """
        if auto_delete is not None and auto_delete not in [1, 2, 3]:
            raise ValueError(f'Unsupported auto-delete option: {auto_delete}')
        self.driver.click(NEW_MESSAGE_STATE_NEW_GROUP_BUTTON)
        for member in members:
            self.gtl_logger.info(f'adding group member {member}')
            self.driver.click(NEW_GROUP_MEMBER.format(member=member))
        self.driver.click(NEW_GROUP_NEXT)
        self.gtl_logger.info(f'Entering group name "{group_name}" and setting auto-delete option')
        self.driver.send_keys(NEW_GROUP_NAME_INPUT, group_name)
        if auto_delete:
            self.driver.click(NEW_GROUP_AUTO_DELETE_OPTION)
            self.driver.click(f'//android.widget.LinearLayout/android.widget.FrameLayout[{auto_delete}]')
        self.driver.click(NEW_GROUP_DONE_BUTTON)

    @action(chat_settings_state)
    def add_members(self, new_members: list[str], conversation: str):
        """
        Adds members to a given chat group.

        :param new_members: a list of members to add to the group.
        :param conversation: the chat group name to add the members to.
        :raises PumaClickException: If not a valid group chat, or if the current user does not have permissions to
        add members.
        """
        if not Telegram.chat_settings_state.can_add_members(self.driver):
            raise PumaClickException(f"Cannot add members in conversation {conversation}. "
                                     f"Check whether it is a group and whether permissions are setup correctly.")
        self.driver.click(CHAT_SETTINGS_STATE_ADD_MEMBERS)
        for name in new_members:
            self.driver.send_keys(CHAT_SETTINGS_STATE_ADD_MEMBERS_SEARCH, name)
            try:
                self.driver.click(
                    f'//androidx.recyclerview.widget.RecyclerView//android.widget.TextView[lower-case(@text)=lower-case("{name}")]')
            except PumaClickException:
                raise PumaClickException(
                    f'Could not add member {name} as it could not be found in the UI after a search')
        self.driver.click('//android.widget.ImageView[@content-desc="Next"]')
        self.driver.click('//android.widget.TextView[@text="Add"]')

    @action(chat_settings_state)
    def remove_member(self, member: str, conversation: str = None):
        """
        Removes a member from a chat group.

        :param member: the member to remove.
        :param conversation: the group chat to remove the user from. Not needed when already in a group chat.
        :raises PumaClickException: If not a valid group chat, or if the current user does not have permissions to
        remove members.
        """
        if not Telegram.chat_settings_state.can_add_members(self.driver):
            raise PumaClickException(f"Cannot remove members in conversation {conversation}. "
                                     f"Check whether it is a group and whether permissions are setup correctly.")
        self.driver.long_click_element(CHAT_SETTINGS_STATE_MEMBER_CONTEXT.format(member=member))
        self.driver.click(CHAT_SETTINGS_STATE_REMOVE_MEMBER_BUTTON)

    @action(chat_settings_state)
    def edit_group_name(self, conversation: str, new_group_name: str):
        """
        Changes the name of a chat group.

        :param conversation: the chat group to rename
        :param new_group_name: the new chat group name
        :raises PumaClickException: If not a valid group chat, or if the current user does not have permissions to
        rename the group.
        """
        if not Telegram.chat_settings_state.can_add_members(self.driver):
            raise PumaClickException(f"Cannot rename conversation {conversation}. "
                                     f"Check whether it is a group and whether permissions are setup correctly.")
        self.driver.click(CHAT_SETTINGS_EDIT_GROUP_BUTTON)
        self.driver.send_keys(EDIT_GROUP_NAME, new_group_name)
        self.driver.click(EDIT_GROUP_DONE)

    @action(chat_settings_state)
    def edit_group_description(self, conversation: str, description: str):
        """
        Changes the description of a chat group.

        :param conversation: the chat group to rename
        :param description: the new description of the chat group
        :raises PumaClickException: If not a valid group chat, or if the current user does not have permissions to
        add members.
        """
        if not Telegram.chat_settings_state.can_add_members(self.driver):
            raise PumaClickException(f"Cannot rename conversation {conversation}. "
                                     f"Check whether it is a group and whether permissions are setup correctly.")
        self.driver.click(CHAT_SETTINGS_EDIT_GROUP_BUTTON)
        self.driver.send_keys(EDIT_GROUP_DESCRIPTION, description)
        self.driver.click(EDIT_GROUP_DONE)

    @action(chat_state)
    def delete_and_leave_group(self, conversation: str):
        """
        Leaves the group and deletes it from the conversations overview.

        :param conversation: the name of the group chat to leave.
        """
        self.driver.click(CHAT_STATE_THREE_DOTS)
        self.driver.click(CHAT_STATE_DELETE_AND_LEAVE_GROUP)
        if self.driver.is_present(CHAT_STATE_DELETE_AND_LEAVE_GROUP_FOR_ALL):
            self.driver.click(CHAT_STATE_DELETE_AND_LEAVE_GROUP_FOR_ALL)
        self.driver.click(CHAT_STATE_DELETE_AND_LEAVE_GROUP_CONFIRM_BUTTON)

    def _in_voice_message_mode(self):
        return self.driver.get_element(CHAT_STATE_RECORD_VIDEO_OR_AUDIO_MESSAGE).get_attribute(
            'content-desc') == 'Record voice message'



from enum import Enum

from puma.apps.android.google_play_store import logger
from puma.state_graph.action import action
from puma.state_graph.popup_handler import PopUpHandler
from puma.state_graph.puma_driver import supported_version, PumaDriver
from puma.state_graph.state import SimpleState, compose_clicks, ContextualState
from puma.state_graph.state_graph import StateGraph
from puma.state_graph.utils import is_valid_package_name

APPLICATION_PACKAGE = 'com.android.vending'

PAYMENT_POPUP_HANDLER = PopUpHandler(['//android.widget.TextView[@text="Pay your way with Google Play"]',
                                      '//android.widget.ImageView[@content-desc="close"]'],
                                     ['//android.widget.ImageView[@content-desc="close"]'])
LOCAL_RECOMMENDATIONS_POPUP_HANDLER = PopUpHandler(
    ['//android.widget.TextView[@text="Want to see local recommendations in Google Play?"]',
     '//android.widget.Button[@text="No thanks"]'], ['//android.widget.Button[@text="No thanks"]'])
TRY_GOOGLE_PASS_POPUP_HANDLER = PopUpHandler(
    ['//android.widget.TextView[@resource-id="com.android.vending:id/0_resource_name_obfuscated" and @text="Try Google Play Pass"]'],
    ['//android.widget.Button[@resource-id="com.android.vending:id/0_resource_name_obfuscated" and @text="Not now"]'])
GOOGLE_PLAY_POINTS_POPUP_HANDLER = PopUpHandler(
    ['//*[@text="Introducing google play points"]'],
    ['//*[@text="Not now"]'])
COMPLETE_ACCOUNT_POPUP_HANDLER = PopUpHandler(
    ['//*[@text="Complete account setup"]'],
    ['//*[@text="Continue"]', '//*[@text="Skip"]'])

ACCOUNT_ICON = '//android.widget.FrameLayout[starts-with(@content-desc, "Signed in as")]'

HOME_SCREEN_TABS = '(//android.view.View[count(.//android.widget.TextView[@text="Games" or @text="Apps" or @text="Search" or @text="Books"]) = 4])[last()]'
APPS_TAB_SELECTED = '//android.view.View[.//android.widget.ImageView[@selected="true"] and ./android.widget.TextView[@text="Apps"]]'

APP_PAGE_INSTALL_BUTTON = '//android.view.View[@content-desc="Install"]'
APP_PAGE_UNINSTALL_BUTTON = '//android.view.View[@content-desc="Uninstall"]'
APP_PAGE_UNINSTALL_SURE_BUTTON = '//android.view.View[@content-desc="Uninstall"]'
APP_PAGE_UPDATE_BUTTON = '//android.view.View[@content-desc="Update"]'
APP_PAGE_CANCEL_INSTALL_BUTTON = '//android.view.View[@content-desc="Cancel"]'
APP_PAGE_THREE_DOTS = '//android.view.View[@content-desc="More options"]'
APP_PAGE_NAVIGATE_UP = '//android.view.View[@content-desc="Navigate up"]'

PROFILE_GOOGLE = '//android.widget.Button[@resource-id="com.android.vending:id/0_resource_name_obfuscated" and @text="Google Account"]'
UPDATE_ALL_BUTTON = '//android.view.View[@content-desc="Update all"]'
MANAGE_APP_STATE = '//android.widget.TextView[@text="Manage apps and device"]'
MANAGE_APP_STATE_SYNC = '//android.widget.TextView[@text="Sync apps to devices"]'
MANAGE_APPS_AND_DEVICES = '//android.widget.TextView[@resource-id="com.android.vending:id/0_resource_name_obfuscated" and @text="Manage apps and device"]'


class AppState(Enum):
    UNKNOWN = 0
    NOT_INSTALLED = 1
    INSTALLED = 2
    UPDATE_AVAILABLE = 3
    INSTALLING = 4
    INSTALLING_UPDATE = 5


class AppPage(SimpleState, ContextualState):
    """
    A state representing the app page of a specific application.

    This class extends both SimpleState and ContextualState, but the contextual aspect is an outlier. See the
    validate_context method.
    """
    def __init__(self, parent_state):
        """
        Initializes the App page state with a given parent state.

        :param parent_state: The parent state of this app page state.
        """
        super().__init__(
            xpaths=[HOME_SCREEN_TABS, APP_PAGE_THREE_DOTS],
            parent_state=parent_state,
            parent_state_transition=compose_clicks([APP_PAGE_NAVIGATE_UP], "navigate_up"))
        # keep a dict that tracks which app pages were opened last on which device. See validate_context()
        self.last_opened = {}

    def validate_context(self, driver: PumaDriver, package_name: str = None) -> bool:
        """
        The package name cannot be found in the UI, except by opening the share menu and checking the URL that can be
        shared. We do not want to use this, because that would mean triggering UI actions every time this contextual
        state is verified.
        Therefore, we opted to store the last opened package name for each device in this State. This means that the
        contextual state is not actually verified against the UI, but against the package name that was opened with the
        open_app_page method. When switching between two app pages, this works as long as the user does not interrupt
        Puma during these UI actions.

        :param driver: Puma driver
        :param package_name: package name
        :return boolean
        """
        if not package_name:
            return True
        return self.last_opened[driver.udid] == package_name

    def open_app_page(self, driver: PumaDriver, package_name: str = None):
        """
        Opens the app page for a specific package name. This is done by opening an intent, opening the URL of the app
        page in the play store application.
        This method also stores which package name's app page was opened on which device, enabling contextual validation
        in validate_context().

        :param driver: Puma driver
        :param package_name: package name
        """
        if not is_valid_package_name(package_name):
            raise ValueError(f'Invalid package name: {package_name}')
        driver.open_url(f'https://play.google.com/store/apps/details?id={package_name}')
        self.last_opened[driver.udid] = package_name


@supported_version("48.3.25-31")
class GooglePlayStore(StateGraph):
    """
    A class representing a state graph for managing UI states and transitions in the Google Play Store.

    This class uses a state machine approach to manage transitions between different states
    of the Play store UI. It provides methods to navigate between states, validate states,
    and handle unexpected states or errors.
    """
    apps_tab_state = SimpleState([ACCOUNT_ICON, HOME_SCREEN_TABS, APPS_TAB_SELECTED], initial_state=True)
    profile_state = SimpleState([MANAGE_APPS_AND_DEVICES, PROFILE_GOOGLE], parent_state=apps_tab_state)
    manage_apps_state = SimpleState([MANAGE_APP_STATE, MANAGE_APP_STATE_SYNC], parent_state=apps_tab_state)
    app_page_state = AppPage(parent_state=apps_tab_state)

    apps_tab_state.to(profile_state, compose_clicks([ACCOUNT_ICON], name='click_profile'))
    profile_state.to(manage_apps_state, compose_clicks([MANAGE_APPS_AND_DEVICES], name='click_manage_apps_and_devices'))
    app_page_state.from_states([apps_tab_state, profile_state, manage_apps_state], app_page_state.open_app_page)

    def __init__(self, device_udid):
        """
        Initializes the Google Play Store with a device UDID.

        :param device_udid: The unique device identifier for the Android device.
        """
        StateGraph.__init__(self, device_udid, APPLICATION_PACKAGE)
        self.add_popup_handler(PAYMENT_POPUP_HANDLER)
        self.add_popup_handler(LOCAL_RECOMMENDATIONS_POPUP_HANDLER)
        self.add_popup_handler(TRY_GOOGLE_PASS_POPUP_HANDLER)
        self.add_popup_handler(GOOGLE_PLAY_POINTS_POPUP_HANDLER)
        self.add_popup_handler(COMPLETE_ACCOUNT_POPUP_HANDLER)

    def _get_app_state_internal(self) -> AppState:
        """
        Util method for @action methods on the app_page_state.
        :return the AppState of the application page, based on found UI elements.
        """
        if self.driver.is_present(APP_PAGE_INSTALL_BUTTON):
            return AppState.NOT_INSTALLED
        if self.driver.is_present(APP_PAGE_UPDATE_BUTTON):
            return AppState.UPDATE_AVAILABLE
        if self.driver.is_present(APP_PAGE_UNINSTALL_BUTTON):
            return AppState.INSTALLED
        if self.driver.is_present(APP_PAGE_CANCEL_INSTALL_BUTTON):
            if self.driver.is_present(APP_PAGE_UNINSTALL_BUTTON):
                return AppState.INSTALLING_UPDATE
            else:
                return AppState.INSTALLING
        logger.error(f'Could not determine the install state of the current app.')
        return AppState.UNKNOWN

    @action(app_page_state)
    def get_app_state(self, package_name: str) -> AppState:
        """
        Returns the AppState of the given application. Possible states are:
        INSTALLED, NOT_INSTALLED, UPDATE_AVAILABLE, INSTALLING, INSTALL_UPDATE, and UNKNOWN.
        :return the AppState of the given application
        """
        return self._get_app_state_internal()

    @action(app_page_state)
    def install_app(self, package_name: str = None):
        """
        Installs the given application. If the application is already installed (or is being installed) this method
        will log a warning and do nothing.
        :param package_name: The exact package name of the application.
        """
        if self._get_app_state_internal() != AppState.NOT_INSTALLED:
            self.gtl_logger.warn(f'Tried to install app {package_name}, but it was already installed')
            return
        self.driver.click(APP_PAGE_INSTALL_BUTTON)

    @action(app_page_state)
    def uninstall_app(self, package_name: str = None):
        """
        Uninstalls the given application. If the application is not installed this method will log a warning and do
        nothing.
        :param package_name: The exact package name of the application.
        """
        if self._get_app_state_internal() not in [AppState.INSTALLED, AppState.UPDATE_AVAILABLE]:
            self.gtl_logger.warn(f'Tried to uninstall app {package_name}, but it was not installed')
            return
        self.driver.click(APP_PAGE_UNINSTALL_BUTTON)
        self.driver.click(APP_PAGE_UNINSTALL_SURE_BUTTON)

    @action(app_page_state)
    def update_app(self, package_name: str = None):
        """
        Updates the given application. If no update is available this method will log a warning and do nothing.
        :param package_name: The exact package name of the application.
        """
        if self._get_app_state_internal() != AppState.UPDATE_AVAILABLE:
            self.gtl_logger.warn(f'Tried to update app {package_name}, but there is no update available')
            return
        self.driver.click(APP_PAGE_UPDATE_BUTTON)

    @action(manage_apps_state)
    def update_all_apps(self):
        """
        Updates all applications. If no updates are available this method will log a warning and do nothing.
        """
        if not self.driver.is_present(UPDATE_ALL_BUTTON):
            self.gtl_logger.warn('Tried to update all apps, but update button not visible. All apps are probably up-to-date.')
            return
        self.driver.click(UPDATE_ALL_BUTTON)



import unittest
from unittest.mock import Mock

from puma.state_graph.puma_driver import PumaClickException
from puma.state_graph.action import action
from puma.state_graph.puma_driver import PumaDriver
from puma.state_graph.state import SimpleState, ContextualState, State
from puma.state_graph.state_graph import StateGraph
from puma.utils import gtl_logging


class MockChatState(SimpleState, ContextualState):

    def __init__(self, parent_state: State):
        super().__init__(xpaths=['xpath'], parent_state=parent_state)

    def validate_context(self, driver: PumaDriver, conversation: str = None) -> bool:
        return True


class MockApplication(StateGraph):
    main_state = SimpleState(['xpath'], initial_state=True)
    settings_state = SimpleState(['xpath'], parent_state=main_state)
    chat_state = MockChatState(parent_state=main_state)

    main_state.to(settings_state, lambda: print('opening setting screen'))
    main_state.to(chat_state, lambda: print('opening chat'))

    # don't call super.__init__ so we do not try to connect to a real device
    def __init__(self, driver=Mock(), gtl_logger=Mock()):
        self.current_state = self.initial_state
        self.driver = driver
        self.gtl_logger = gtl_logger

        self.username = None
        self.messages = {}

    @action(settings_state)
    def change_username(self, new_name: str):
        self.username = new_name

    @action(chat_state)
    def send_message(self, message: str, conversation: str = None, ):
        if conversation not in self.messages:
            self.messages = []
        self.messages.append(message)

    def instance_verify_username_equals(self, new_name):
        return self.username == new_name

    @staticmethod
    def static_verify_username_equals(app, new_name):
        return app.username == new_name


class TestVerifyWith(unittest.TestCase):

    def setUp(self):
        # used to capture values inside passed 'verify_with' functions
        self.captures = []

    def test_verify_with_callable_is_called(self):
        # test if the verify_with is called by just capturing a boolean
        def should_be_called():
            self.captures.append(True)
            return True

        application = MockApplication()
        application.change_username(new_name='Capture this', verify_with=should_be_called)

        self.assertEqual(self.captures[0], True)

    def test_verify_with_not_callable_raises_exception(self):
        application = MockApplication()

        with self.assertRaisesRegex(TypeError, "'verify_with' must be a callable"):
            application.change_username(new_name='This is illegal', verify_with='not a callable')

    def test_verify_with_using_static_method(self):
        mock_logger = gtl_logging.create_gtl_logger('mock_udid')

        with self.assertLogs(mock_logger, level='INFO') as logs:
            application = MockApplication(gtl_logger=mock_logger)
            application.change_username(new_name='NewName', verify_with=MockApplication.static_verify_username_equals)

            self.assertIn("INFO:mock_udid:Action 'change_username' succeeded", logs.output)

    def test_verify_with_using_instance_method(self):
        mock_logger = gtl_logging.create_gtl_logger('mock_udid')

        with self.assertLogs(mock_logger, level='INFO') as logs:
            application = MockApplication(gtl_logger=mock_logger)
            application.change_username(new_name='NewName', verify_with=application.instance_verify_username_equals)

            self.assertIn("INFO:mock_udid:Action 'change_username' succeeded", logs.output)

    def test_verify_with_raises_puma_click_exception_is_logged(self):
        def verification_throws_puma_click_exception():
            raise PumaClickException('puma click exception')

        mock_logger = gtl_logging.create_gtl_logger('mock_udid')

        with self.assertLogs(mock_logger, level='INFO') as logs:
            application = MockApplication(gtl_logger=mock_logger)
            application.change_username(new_name='MyName', verify_with=verification_throws_puma_click_exception)

            # assert we logged what happened
            self.assertIn("WARNING:mock_udid:Verifying with 'verification_throws_puma_click_exception' failed due to exception: puma click exception", logs.output)
            # assert the action happened, i.e. our username has changed
            self.assertEqual(application.username, 'MyName')

    def test_verify_with_raises_generic_exception_is_propagated(self):
        def verification_throws_puma_click_exception():
            raise Exception('generic exception')

        application = MockApplication()

        with self.assertRaisesRegex(Exception, 'generic exception'):
            application.change_username(new_name='MyName', verify_with=verification_throws_puma_click_exception)

    def test_state_graph_is_passed_to_verify_with(self):
        # test if the verify_with receives the app by capturing it inside the function
        def should_receive_app(app):
            self.captures.append(app)
            return True

        application = MockApplication()
        application.change_username(new_name='ignore', verify_with=should_receive_app)

        self.assertIs(self.captures[0], application)

    def test_app_and_context_are_passed_to_verify_with(self):
        # test if the verify_with receives both by capturing them inside the function
        def should_receive_both(app, conversation):
            self.captures.append(app)
            self.captures.append(conversation)
            return True

        application = MockApplication()
        application.send_message(conversation='Bob', message='ignore', verify_with=should_receive_both)

        self.assertIs(self.captures[0], application)
        self.assertEqual(self.captures[1], 'Bob')

    def test_app_and_context_and_argument_are_passed_to_verify_with(self):
        # test if the verify_with receives all by capturing them inside the function
        def should_receive_all(app, conversation, message):
            self.captures.append(app)
            self.captures.append(conversation)
            self.captures.append(message)
            return True

        application = MockApplication()
        application.send_message(conversation='Bob', message='Hello', verify_with=should_receive_all)

        # our verify_with should have received the conversation
        self.assertIs(self.captures[0], application)
        self.assertEqual(self.captures[1], 'Bob')
        self.assertEqual(self.captures[2], 'Hello')

    def test_verify_with_logs_and_succeeds(self):
        def log_warning_and_return_true(app):
            gtl_logger = app.gtl_logger
            gtl_logger.info('post verification succeeded')
            return True

        mock_logger = gtl_logging.create_gtl_logger('mock_udid')

        with self.assertLogs(mock_logger, level='INFO') as logs:
            application = MockApplication(gtl_logger=mock_logger)
            application.change_username(new_name='MyName', verify_with=log_warning_and_return_true)

            # assert we called the gtl_logger we passed to our application, and it logged the expected message
            self.assertIn('INFO:mock_udid:post verification succeeded', logs.output)
            # assert the action happened, i.e. our username has changed
            self.assertEqual(application.username, 'MyName')

    def test_verify_with_logs_and_fails_still_executed_action(self):
        # failed verification should still result in the action being executed

        def log_warning_and_return_false(app):
            app.gtl_logger.warn('post verification failed')
            return False

        mock_logger = gtl_logging.create_gtl_logger('mock_udid')

        with self.assertLogs(mock_logger, level='WARN') as logs:
            application = MockApplication(gtl_logger=mock_logger)
            application.change_username(new_name='NewName', verify_with=log_warning_and_return_false)

            # assert we called the gtl_logger we passed to our application, and it logged the expected message
            self.assertIn('WARNING:mock_udid:post verification failed', logs.output)
            # assert the action still happened, i.e. our username has changed
            self.assertEqual(application.username, 'NewName')

    def test_verify_with_logs_and_fails_still_executes_following_action(self):
        # failed verification should still result in follow-up actions being executed

        def log_warning_and_return_false(app):
            app.gtl_logger.warn('post verification failed')
            return False

        mock_logger = gtl_logging.create_gtl_logger('mock_udid')

        with self.assertLogs(mock_logger, level='WARN') as logs:
            application = MockApplication(gtl_logger=mock_logger)
            application.change_username(new_name='OldName', verify_with=log_warning_and_return_false)
            application.change_username(new_name='NewName')

            # assert we called the gtl_logger we passed to our application, and it logged the expected message
            self.assertIn('WARNING:mock_udid:post verification failed', logs.output)
            # assert the follow up action still happened, i.e. our username has changed
            self.assertEqual(application.username, 'NewName')

    def test_verify_with_can_change_states_but_framework_returns_to_simple_state_before(self):
        # inside we move away from the settings_state simple state
        def change_states_and_return():
            application.go_to_state(application.main_state)
            return True

        application = MockApplication()
        application.change_username(new_name='ignore', verify_with=change_states_and_return)

        self.assertIs(application.current_state, application.settings_state)

    def test_verify_with_can_change_states_but_framework_returns_to_contextual_state_before(self):
        # inside we move away from the chat contextual state
        def change_states_and_return():
            application.go_to_state(application.main_state)
            return True

        application = MockApplication()
        application.send_message(conversation='ignore', message='ignore', verify_with=change_states_and_return)

        self.assertIs(application.current_state, application.chat_state)

    def test_verify_with_can_change_states_and_manually_returns_to_contextual_state_before(self):
        # inside we move away and back to the chat contextual state
        def change_states_and_return():
            application.go_to_state(application.main_state)
            application.go_to_state(application.chat_state)
            return True

        application = MockApplication()
        application.send_message(conversation='ignore', message='ignore', verify_with=change_states_and_return)

        self.assertIs(application.current_state, application.chat_state)

    def test_action_decorated_function_cant_have_parameter_named_verify_with(self):
        @action(MockApplication.settings_state)
        def this_should_throw_exception(verify_with):
            pass

        with self.assertRaisesRegex(Exception, "can't contain a parameter named 'verify_with'"):
            this_should_throw_exception()



import inspect
import re

from puma.state_graph import logger
from puma.state_graph.puma_driver import PumaClickException


def safe_func_call(func, **kwargs):
    """
    Safely calls a function with the provided keyword arguments.

    This function filters the provided keyword arguments to only include those that are
    defined in the function's signature. It then attempts to call the function with these
    filtered arguments. If a PumaClickException occurs during the function call, it catches
    the exception, prints an error message, and returns None.

    :param func: The function to be called.
    :param kwargs: Arbitrary keyword arguments to pass to the function.
    :return: The result of the function call, or None if an exception occurs.
    """
    bound_args = filter_arguments(func, **kwargs)
    try:
        return func(**bound_args.arguments)
    except PumaClickException as pce:
        logger.warning(f"A problem occurred during a safe function call, recovering.. {pce}")
        return None


def filter_arguments(func, **kwargs) -> inspect.BoundArguments:
    signature = inspect.signature(func)
    filtered_args = {
        k: v for k, v in kwargs.items() if k in signature.parameters
    }
    bound_args = signature.bind(**filtered_args)
    bound_args.apply_defaults()
    return bound_args

def is_valid_package_name(package_name: str) -> bool:
    pattern = r'^[a-z_][a-z0-9_]+(\.[a-z_][a-z0-9_]*)*$'
    return bool(re.fullmatch(pattern, package_name)) and len(package_name) <= 100